"""
Auto Classfier for Heterogeneous Node Classification
"""
import time
import json

from copy import deepcopy
from typing import Sequence

import torch
import numpy as np
import yaml

from ..base import BaseClassifier
from ...base import _parse_hp_space, _initialize_single_model, _parse_model_hp
from ....module.train import TRAINER_DICT, BaseNodeClassificationHetTrainer
from ....module.train import get_feval
from ...utils import LeaderBoard, set_seed
from ....utils import get_logger

LOGGER = get_logger("HeteroNodeClassifier")

class AutoHeteroNodeClassifier(BaseClassifier):
    """
    Auto Multi-class HeteroGraph Node Classifier.

    Used to automatically solve the heterogeneous node classification problems.

    Parameters
    ----------
    feature_module: autogl.module.feature.BaseFeatureEngineer or str or None
        The (name of) auto feature engineer used to process the given dataset. Default ``deepgl``.
        Disable feature engineer by setting it to ``None``.

    graph_models: list of autogl.module.model.BaseModel or list of str
        The (name of) models to be optimized as backbone. Default ``['gat', 'gcn']``.

    hpo_module: autogl.module.hpo.BaseHPOptimizer or str or None
        The (name of) hpo module used to search for best hyper parameters. Default ``anneal``.
        Disable hpo by setting it to ``None``.

    ensemble_module: autogl.module.ensemble.BaseEnsembler or str or None
        The (name of) ensemble module used to ensemble the multi-models found. Default ``voting``.
        Disable ensemble by setting it to ``None``.

    max_evals: int (Optional)
        If given, will set the number eval times the hpo module will use.
        Only be effective when hpo_module is ``str``. Default ``None``.

    default_trainer: str (Optional)
        The (name of) the trainer used in this solver. Default to ``NodeClassificationFull``.

    trainer_hp_space: list of dict (Optional)
        trainer hp space or list of trainer hp spaces configuration.
        If a single trainer hp is given, will specify the hp space of trainer for every model.
        If a list of trainer hp is given, will specify every model with corrsponding
        trainer hp space.
        Default ``None``.

    model_hp_spaces: list of list of dict (Optional)
        model hp space configuration.
        If given, will specify every hp space of every passed model. Default ``None``.
        If the encoder(-decoder) is passed, the space should be a dict containing keys "encoder"
        and "decoder", specifying the detailed encoder decoder hp spaces.

    size: int (Optional)
        The max models ensemble module will use. Default ``None``.

    device: torch.device or str
        The device where model will be running on. If set to ``auto``, will use gpu when available.
        You can also specify the device by directly giving ``gpu`` or ``cuda:0``, etc.
        Default ``auto``.
    """

    def __init__(
        self,
        graph_models=("han", "hgt"),
        hpo_module="anneal",
        ensemble_module="voting",
        max_evals=50,
        default_trainer="NodeClassificationHet",
        trainer_hp_space=None,
        model_hp_spaces=None,
        size=4,
        device="auto",
    ):

        super().__init__(
            # currently we do not support feature engineering
            feature_module=None,
            graph_models=graph_models,
            # currently we do not support nas for heterogeneous node classifier
            nas_algorithms=None,
            nas_spaces=None,
            nas_estimators=None,
            hpo_module=hpo_module,
            ensemble_module=ensemble_module,
            max_evals=max_evals,
            default_trainer=default_trainer,
            trainer_hp_space=trainer_hp_space,
            model_hp_spaces=model_hp_spaces,
            size=size,
            device=device,
        )

        # data to be kept when fit
        self.dataset = None

    def _init_graph_module(
        self, graph_models, num_classes, num_features, feval, device, loss, dataset
    ) -> "AutoHeteroNodeClassifier":
        # load graph network module
        self.graph_model_list = []

        for i, model in enumerate(graph_models):
            # init the trainer
            if not isinstance(model, BaseNodeClassificationHetTrainer):
                trainer = (
                    self._default_trainer if not isinstance(self._default_trainer, (tuple, list))
                    else self._default_trainer[i]
                )
                if isinstance(trainer, str):
                    if trainer not in TRAINER_DICT:
                        raise KeyError(f"Does not support trainer {trainer}")
                    trainer = TRAINER_DICT[trainer]()
                if isinstance(model, (tuple, list)):
                    trainer.encoder = model[0]
                    trainer.decoder = model[1]
                else:
                    trainer.encoder = model
            else:
                trainer = model

            # set model hp space
            if self._model_hp_spaces is not None:
                if self._model_hp_spaces[i] is not None:
                    if isinstance(self._model_hp_spaces[i], dict):
                        encoder_hp_space = self._model_hp_spaces[i].get('encoder', None)
                        decoder_hp_space = self._model_hp_spaces[i].get('decoder', None)
                    else:
                        encoder_hp_space = self._model_hp_spaces[i]
                        decoder_hp_space = None
                    if encoder_hp_space is not None:
                        trainer.encoder.hyper_parameter_space = encoder_hp_space
                    if decoder_hp_space is not None:
                        trainer.decoder.hyper_parameter_space = decoder_hp_space
            
            # set trainer hp space
            if self._trainer_hp_space is not None:
                if isinstance(self._trainer_hp_space[0], list):
                    current_hp_for_trainer = self._trainer_hp_space[i]
                else:
                    current_hp_for_trainer = self._trainer_hp_space
                trainer.hyper_parameter_space = current_hp_for_trainer

            trainer.num_features = num_features
            trainer.num_classes = num_classes
            trainer.from_dataset(dataset)
            trainer.loss = loss
            trainer.feval = feval
            trainer.to(device)
            self.graph_model_list.append(trainer)

        return self

    # pylint: disable=arguments-differ
    def fit(
        self,
        dataset,
        time_limit=-1,
        evaluation_method="infer",
        seed=None,
    ) -> "AutoHeteroNodeClassifier":
        """
        Fit current solver on given dataset.

        Parameters
        ----------
        dataset: autogl.data.Dataset
            The dataset needed to fit on. This dataset must have only one graph.

        time_limit: int
            The time limit of the whole fit process (in seconds). If set below 0,
            will ignore time limit. Default ``-1``.

        inplace: bool
            Whether we process the given dataset in inplace manner. Default ``False``.
            Set it to True if you want to save memory by modifying the given dataset directly.

        evaluation_method: (list of) str or autogl.module.train.evaluation
            A (list of) evaluation method for current solver. If ``infer``, will automatically
            determine. Default ``infer``.

        seed: int (Optional)
            The random seed. If set to ``None``, will run everything at random.
            Default ``None``.

        Returns
        -------
        self: autogl.solver.AutoNodeClassifier
            A reference of current solver.
        """
        set_seed(seed)

        if time_limit < 0:
            time_limit = 3600 * 24
        time_begin = time.time()

        graph_data = dataset[0]
        field = dataset.schema["target_node_type"]
        all_labels = graph_data.nodes[field].data['label']
        num_classes = all_labels.max().item() + 1

        # initialize leaderboard
        if evaluation_method == "infer":
            if hasattr(dataset, "metric"):
                evaluation_method = [dataset.metric]
            else:
                num_of_label = num_classes
                if num_of_label == 2:
                    evaluation_method = ["auc"]
                else:
                    evaluation_method = ["acc"]
        assert isinstance(evaluation_method, list)
        evaluator_list = get_feval(evaluation_method)

        self.leaderboard = LeaderBoard(
            [e.get_eval_name() for e in evaluator_list],
            {e.get_eval_name(): e.is_higher_better() for e in evaluator_list},
        )

        # set up the dataset
        assert ("train_mask" in graph_data.nodes[field].data 
            and "val_mask" in graph_data.nodes[field].data), ("Currently only support"
            " Dataset with default train/val/test split")

        self.dataset = dataset

        # check whether the dataset has features.
        # currently we only support hetero graph classification with features.

        feat = graph_data.nodes[field].data['feat']
        assert feat is not None, (
            "Does not support fit on non node-feature dataset!"
            " Please add node features to dataset or specify feature engineers that generate"
            " node features."
        )

        num_features = feat.size(-1)

        # initialize graph networks
        self._init_graph_module(
            self.gml,
            num_features=num_features,
            num_classes=num_classes,
            feval=evaluator_list,
            device=self.runtime_device,
            loss="nll_loss" if not hasattr(dataset, "loss") else self.dataset.loss,
            dataset=dataset
        )

        # train the models and tune hpo
        result_valid = []
        names = []
        for idx, model in enumerate(self.graph_model_list):
            time_for_each_model = (time_limit - time.time() + time_begin) / (
                len(self.graph_model_list) - idx
            )
            if self.hpo_module is None:
                model.initialize()
                model.train(dataset, True)
                optimized = model
            else:
                optimized, _ = self.hpo_module.optimize(
                    trainer=model, dataset=dataset, time_limit=time_for_each_model
                )
            # to save memory, all the trainer derived will be mapped to cpu
            optimized.to(torch.device("cpu"))
            name = str(optimized) + "_idx%d" % (idx)
            names.append(name)
            performance_on_valid, _ = optimized.get_valid_score(return_major=False)
            result_valid.append(optimized.get_valid_predict_proba().cpu().numpy())
            self.leaderboard.insert_model_performance(
                name,
                dict(
                    zip(
                        [e.get_eval_name() for e in evaluator_list],
                        performance_on_valid,
                    )
                ),
            )
            self.trained_models[name] = optimized

        # fit the ensemble model
        if self.ensemble_module is not None:
            performance = self.ensemble_module.fit(
                result_valid,
                all_labels[graph_data.nodes[field].data["val_mask"]].cpu().numpy(),
                names,
                evaluator_list,
                n_classes=num_classes,
            )
            self.leaderboard.insert_model_performance(
                "ensemble",
                dict(zip([e.get_eval_name() for e in evaluator_list], performance)),
            )

        return self

    def fit_predict(
        self,
        dataset,
        time_limit=-1,
        evaluation_method="infer",
        use_ensemble=True,
        use_best=True,
        name=None,
    ) -> np.ndarray:
        """
        Fit current solver on given dataset and return the predicted value.

        Parameters
        ----------
        dataset: torch_geometric.data.dataset.Dataset
            The dataset needed to fit on. This dataset must have only one graph.

        time_limit: int
            The time limit of the whole fit process (in seconds).
            If set below 0, will ignore time limit. Default ``-1``.

        inplace: bool
            Whether we process the given dataset in inplace manner. Default ``False``.
            Set it to True if you want to save memory by modifying the given dataset directly.

        train_split: float or int (Optional)
            The train ratio (in ``float``) or number (in ``int``) of dataset. If you want to
            use default train/val/test split in dataset, please set this to ``None``.
            Default ``None``.

        val_split: float or int (Optional)
            The validation ratio (in ``float``) or number (in ``int``) of dataset. If you want
            to use default train/val/test split in dataset, please set this to ``None``.
            Default ``None``.

        balanced: bool
            Wether to create the train/valid/test split in a balanced way.
            If set to ``True``, the train/valid will have the same number of different classes.
            Default ``False``.

        evaluation_method: (list of) str or autogl.module.train.evaluation
            A (list of) evaluation method for current solver. If ``infer``, will automatically
            determine. Default ``infer``.

        use_ensemble: bool
            Whether to use ensemble to do the predict. Default ``True``.

        use_best: bool
            Whether to use the best single model to do the predict. Will only be effective when
            ``use_ensemble`` is ``False``.
            Default ``True``.

        name: str or None
            The name of model used to predict. Will only be effective when ``use_ensemble`` and
            ``use_best`` both are ``False``.
            Default ``None``.

        Returns
        -------
        result: np.ndarray
            An array of shape ``(N,)``, where ``N`` is the number of test nodes. The prediction
            on given dataset.
        """
        self.fit(
            dataset=dataset,
            time_limit=time_limit,
            evaluation_method=evaluation_method,
        )
        return self.predict(
            dataset=dataset,
            use_ensemble=use_ensemble,
            use_best=use_best,
            name=name,
        )

    def predict_proba(
        self,
        dataset=None,
        use_ensemble=True,
        use_best=True,
        name=None,
        mask="test",
    ) -> np.ndarray:
        """
        Predict the node probability.

        Parameters
        ----------
        dataset: torch_geometric.data.dataset.Dataset or None
            The dataset needed to predict. If ``None``, will use the processed dataset passed
            to ``fit()`` instead. Default ``None``.

        inplaced: bool
            Whether the given dataset is processed. Only be effective when ``dataset``
            is not ``None``. If you pass the dataset to ``fit()`` with ``inplace=True``, and
            you pass the dataset again to this method, you should set this argument to ``True``.
            Otherwise ``False``. Default ``False``.

        inplace: bool
            Whether we process the given dataset in inplace manner. Default ``False``. Set it to
            True if you want to save memory by modifying the given dataset directly.

        use_ensemble: bool
            Whether to use ensemble to do the predict. Default ``True``.

        use_best: bool
            Whether to use the best single model to do the predict. Will only be effective when
            ``use_ensemble`` is ``False``. Default ``True``.

        name: str or None
            The name of model used to predict. Will only be effective when ``use_ensemble`` and
            ``use_best`` both are ``False``. Default ``None``.

        mask: str
            The data split to give prediction on. Default ``test``.

        Returns
        -------
        result: np.ndarray
            An array of shape ``(N,C,)``, where ``N`` is the number of test nodes and ``C`` is
            the number of classes. The prediction on given dataset.
        """
        if dataset is None:
            dataset = self.dataset
            assert dataset is not None, (
                "Please execute fit() first before" " predicting on remembered dataset"
            )

        if use_ensemble:
            LOGGER.info("Ensemble argument on, will try using ensemble model.")

        if not use_ensemble and use_best:
            LOGGER.info(
                "Ensemble argument off and best argument on, will try using best model."
            )

        if (use_ensemble and self.ensemble_module is not None) or (
            not use_best and name == "ensemble"
        ):
            # we need to get all the prediction of every model trained
            predict_result = []
            names = []
            for model_name in self.trained_models:
                predict_result.append(
                    self._predict_proba_by_name(dataset, model_name, mask)
                )
                names.append(model_name)
            return self.ensemble_module.ensemble(predict_result, names)

        if use_ensemble and self.ensemble_module is None:
            LOGGER.warning(
                "Cannot use ensemble because no ensebmle module is given."
                " Will use best model instead."
            )

        if use_best or (use_ensemble and self.ensemble_module is None):
            # just return the best model we have found
            name = self.leaderboard.get_best_model()
            return self._predict_proba_by_name(dataset, name, mask)

        if name is not None:
            # return model performance by name
            return self._predict_proba_by_name(dataset, name, mask)

        LOGGER.error(
            "No model name is given while ensemble and best arguments are off."
        )
        raise ValueError(
            "You need to specify a model name if you do not want use ensemble and best model."
        )

    def _predict_proba_by_name(self, dataset, name, mask="test"):
        self.trained_models[name].to(self.runtime_device)
        predicted = (
            self.trained_models[name].predict_proba(dataset, mask=mask).cpu().numpy()
        )
        self.trained_models[name].to(torch.device("cpu"))
        return predicted

    def predict(
        self,
        dataset=None,
        use_ensemble=True,
        use_best=True,
        name=None,
        mask="test",
    ) -> np.ndarray:
        """
        Predict the node class number.

        Parameters
        ----------
        dataset: torch_geometric.data.dataset.Dataset or None
            The dataset needed to predict. If ``None``, will use the processed dataset passed
            to ``fit()`` instead. Default ``None``.

        inplaced: bool
            Whether the given dataset is processed. Only be effective when ``dataset``
            is not ``None``. If you pass the dataset to ``fit()`` with ``inplace=True``,
            and you pass the dataset again to this method, you should set this argument
            to ``True``. Otherwise ``False``. Default ``False``.

        inplace: bool
            Whether we process the given dataset in inplace manner. Default ``False``.
            Set it to True if you want to save memory by modifying the given dataset directly.

        use_ensemble: bool
            Whether to use ensemble to do the predict. Default ``True``.

        use_best: bool
            Whether to use the best single model to do the predict. Will only be effective
            when ``use_ensemble`` is ``False``. Default ``True``.

        name: str or None
            The name of model used to predict. Will only be effective when ``use_ensemble``
            and ``use_best`` both are ``False``. Default ``None``.

        mask: str
            The data split to give prediction on. Default ``test``.

        Returns
        -------
        result: np.ndarray
            An array of shape ``(N,)``, where ``N`` is the number of test nodes.
            The prediction on given dataset.
        """
        proba = self.predict_proba(
            dataset, use_ensemble, use_best, name, mask
        )
        return np.argmax(proba, axis=1)


    def evaluate(self, dataset=None,
        use_ensemble=True,
        use_best=True,
        name=None,
        mask="test",
        label=None,
        metric="acc"
    ):
        """
        Evaluate the given dataset.


        Parameters
        ----------
        dataset: torch_geometric.data.dataset.Dataset or None
            The dataset needed to predict. If ``None``, will use the processed dataset passed
            to ``fit()`` instead. Default ``None``.

        inplaced: bool
            Whether the given dataset is processed. Only be effective when ``dataset``
            is not ``None``. If you pass the dataset to ``fit()`` with ``inplace=True``, and
            you pass the dataset again to this method, you should set this argument to ``True``.
            Otherwise ``False``. Default ``False``.

        inplace: bool
            Whether we process the given dataset in inplace manner. Default ``False``. Set it to
            True if you want to save memory by modifying the given dataset directly.

        use_ensemble: bool
            Whether to use ensemble to do the predict. Default ``True``.

        use_best: bool
            Whether to use the best single model to do the predict. Will only be effective when
            ``use_ensemble`` is ``False``. Default ``True``.

        name: str or None
            The name of model used to predict. Will only be effective when ``use_ensemble`` and
            ``use_best`` both are ``False``. Default ``None``.

        mask: str
            The data split to give prediction on. Default ``test``.

        label: torch.Tensor (Optional)
            The groud truth label of the given predicted dataset split. If not passed, will extract
            labels from the input dataset.
        
        metric: str
            The metric to be used for evaluating the model. Default ``acc``.

        Returns
        -------
        score(s): (list of) evaluation scores
            the evaluation results according to the evaluator passed.

        """
        predicted = self.predict_proba(dataset, use_ensemble, use_best, name, mask)
        if dataset is None:
            dataset = self.dataset
        if label is None:
            graph_nodes = dataset[0].nodes[dataset.schema["target_node_type"]].data
            if mask in ["train", "val", "test"]:
                mask = graph_nodes[f"{mask}_mask"]
            label = graph_nodes["label"][mask].cpu().numpy()
        evaluator = get_feval(metric)
        if isinstance(evaluator, Sequence):
            return [evals.evaluate(predicted, label) for evals in evaluator]
        return evaluator.evaluate(predicted, label)

    @classmethod
    def from_config(cls, path_or_dict, filetype="auto") -> "AutoHeteroNodeClassifier":
        """
        Load solver from config file.

        You can use this function to directly load a solver from predefined config dict
        or config file path. Currently, only support file type of ``json`` or ``yaml``,
        if you pass a path.

        Parameters
        ----------
        path_or_dict: str or dict
            The path to the config file or the config dictionary object

        filetype: str
            The filetype the given file if the path is specified. Currently only support
            ``json`` or ``yaml``. You can set to ``auto`` to automatically detect the file
            type (from file name). Default ``auto``.

        Returns
        -------
        solver: autogl.solver.AutoGraphClassifier
            The solver that is created from given file or dictionary.
        """
        assert filetype in ["auto", "yaml", "json"], (
            "currently only support yaml file or json file type, but get type "
            + filetype
        )
        if isinstance(path_or_dict, str):
            if filetype == "auto":
                if path_or_dict.endswith(".yaml") or path_or_dict.endswith(".yml"):
                    filetype = "yaml"
                elif path_or_dict.endswith(".json"):
                    filetype = "json"
                else:
                    LOGGER.error(
                        "cannot parse the type of the given file name, "
                        "please manually set the file type"
                    )
                    raise ValueError(
                        "cannot parse the type of the given file name, "
                        "please manually set the file type"
                    )
            if filetype == "yaml":
                path_or_dict = yaml.load(
                    open(path_or_dict, "r").read(), Loader=yaml.FullLoader
                )
            else:
                path_or_dict = json.load(open(path_or_dict, "r"))

        path_or_dict = deepcopy(path_or_dict)
        solver = cls(None, [], None, None)
        
        models = path_or_dict.pop("models", [{"name": "hgt"}, {"name": "han"}])
        # models should be a list of model
        # with each element in two cases
        # * a dict describing a certain model
        # * a dict containing {"encoder": encoder, "decoder": decoder}
        model_hp_space = [
            _parse_model_hp(model) for model in models
        ]
        model_list = [
            _initialize_single_model(model) for model in models
        ]

        trainer = path_or_dict.pop("trainer", None)
        default_trainer = "NodeClassificationHet"
        trainer_space = None
        if isinstance(trainer, dict):
            # global default
            default_trainer = trainer.pop("name", "NodeClassificationHet")
            trainer_space = _parse_hp_space(trainer.pop("hp_space", None))
            default_kwargs = {"num_features": None, "num_classes": None}
            default_kwargs.update(trainer)
            default_kwargs["init"] = False
            for i in range(len(model_list)):
                model = model_list[i]
                trainer_wrap = TRAINER_DICT[default_trainer](
                    model=model, **default_kwargs
                )
                model_list[i] = trainer_wrap
        elif isinstance(trainer, list):
            # sequential trainer definition
            assert len(trainer) == len(
                model_list
            ), "The number of trainer and model does not match"
            trainer_space = []
            for i in range(len(model_list)):
                train, model = trainer[i], model_list[i]
                default_trainer = train.pop("name", "NodeClassificationHet")
                trainer_space.append(_parse_hp_space(train.pop("hp_space", None)))
                default_kwargs = {"num_features": None, "num_classes": None}
                default_kwargs.update(train)
                default_kwargs["init"] = False
                trainer_wrap = TRAINER_DICT[default_trainer](
                    model=model, **default_kwargs
                )
                model_list[i] = trainer_wrap

        solver.set_graph_models(
            model_list, default_trainer, trainer_space, model_hp_space
        )

        hpo_dict = path_or_dict.pop("hpo", {"name": "anneal"})
        if hpo_dict is not None:
            name = hpo_dict.pop("name")
            solver.set_hpo_module(name, **hpo_dict)

        ensemble_dict = path_or_dict.pop("ensemble", {"name": "voting"})
        if ensemble_dict is not None:
            name = ensemble_dict.pop("name")
            solver.set_ensemble_module(name, **ensemble_dict)

        return solver



# Modified from deeprobust Library
import scipy.sparse as sp
import numpy as np
from numba import njit
@njit
def dropedge_jaccard(A, iA, jA, features, threshold):
    removed_cnt = 0
    for row in range(len(iA)-1):
        for i in range(iA[row], iA[row+1]):
            # print(row, jA[i], A[i])
            n1 = row
            n2 = jA[i]
            a, b = features[n1], features[n2]
            intersection = np.count_nonzero(a*b)
            J = intersection * 1.0 / (np.count_nonzero(a) + np.count_nonzero(b) - intersection)

            if J < threshold:
                A[i] = 0
                # A[n2, n1] = 0
                removed_cnt += 1
    return removed_cnt


@njit
def dropedge_cosine(A, iA, jA, features, threshold):
    removed_cnt = 0
    for row in range(len(iA)-1):
        for i in range(iA[row], iA[row+1]):
            # print(row, jA[i], A[i])
            n1 = row
            n2 = jA[i]
            a, b = features[n1], features[n2]
            inner_product = (a * b).sum()
            C = inner_product / (np.sqrt(np.square(a).sum()) * np.sqrt(np.square(b).sum()) + 1e-8)

            if C < threshold:
                A[i] = 0
                # A[n2, n1] = 0
                removed_cnt += 1
    return removed_cnt

@njit
def dropedge_dis(A, iA, jA, features, threshold):
    removed_cnt = 0
    for row in range(len(iA)-1):
        for i in range(iA[row], iA[row+1]):
            # print(row, jA[i], A[i])
            n1 = row
            n2 = jA[i]
            C = np.linalg.norm(features[n1] - features[n2])
            if C > threshold:
                A[i] = 0
                # A[n2, n1] = 0
                removed_cnt += 1

    return removed_cnt

@njit
def dropedge_both(A, iA, jA, features, threshold1=2.5, threshold2=0.01):
    removed_cnt = 0
    for row in range(len(iA)-1):
        for i in range(iA[row], iA[row+1]):
            # print(row, jA[i], A[i])
            n1 = row
            n2 = jA[i]
            C1 = np.linalg.norm(features[n1] - features[n2])

            a, b = features[n1], features[n2]
            inner_product = (a * b).sum()
            C2 = inner_product / (np.sqrt(np.square(a).sum() + np.square(b).sum())+ 1e-6)
            if C1 > threshold1 or threshold2 < 0:
                A[i] = 0
                # A[n2, n1] = 0
                removed_cnt += 1

    return removed_cnt

def drop_dissimilar_edges(features, adj,threshold,binary_feature=True,metric='similarity'):
    """Drop dissimilar edges.(Faster version using numba)
    """
    if not sp.issparse(adj):
        adj = sp.csr_matrix(adj)

    adj_triu = sp.triu(adj, format='csr')

    if sp.issparse(features):
        features = features.todense().A # make it easier for njit processing

    if metric == 'distance':
        removed_cnt = dropedge_dis(adj_triu.data, adj_triu.indptr, adj_triu.indices, features, threshold=threshold)
    else:
        if binary_feature:
            removed_cnt = dropedge_jaccard(adj_triu.data, adj_triu.indptr, adj_triu.indices, features, threshold=threshold)
        else:
            removed_cnt = dropedge_cosine(adj_triu.data, adj_triu.indptr, adj_triu.indices, features, threshold=threshold)
    print('removed %s edges in the original graph' % removed_cnt)
    modified_adj = adj_triu + adj_triu.transpose()
    return modified_adj

from .._data_preprocessor_registry import DataPreprocessorUniversalRegistry
from ._structure_engineer import StructureEngineer,get_feature,get_edges,set_edges,to_adjacency_list,to_adjacency_matrix,LOGGER
import torch
@DataPreprocessorUniversalRegistry.register_data_preprocessor("gcnjaccard")
class GCNJaccard(StructureEngineer):
    """
    GCNJaccard preprocesses input graph via droppining dissimilar
    edges. See more details in
    Adversarial Examples on Graph Data: Deep Insights into Attack and Defense,
    https://arxiv.org/pdf/1903.01610.pdf.
    """
    def __init__(self, threshold=0.01,binary_feature=True, *args, **kwargs):
        """ drop dissimilar edges with similarity smaller than given threshold

        Parameters
        ----------
        threshold : float
            similarity threshold for dropping edges. If two connected nodes with similarity smaller than threshold, the edge between them will be removed.
        """
        super(GCNJaccard, self).__init__(*args, **kwargs)
        self.threshold=threshold
        self.binary_feature=binary_feature
    def _transform(self,data):
        features = get_feature(data)
        adj = get_edges(data) # edge list 
        LOGGER.info(f'before modified: {adj.shape}')
        adj = to_adjacency_matrix(adj).numpy() # adjacency matrix
        modified_adj = drop_dissimilar_edges(features, adj, self.threshold,self.binary_feature).toarray() # adjacency matrix
        modified_adj = to_adjacency_list(torch.Tensor(modified_adj)) # edge list
        LOGGER.info(f'after modified: {modified_adj.shape}' )
        set_edges(data,modified_adj)
        return data


import torch.utils.data
import typing as _typing


class TargetDependantSampledData:
    """
    A uniform aggregation of sampled data for one mini-batch,
    generally sampler by target-dependent sampler.
    Node-wise Sampling and Layer-wise Sampling techniques are definitely target-dependent,
    for which each sampled subgraph depends on the corresponding target nodes.
    Besides, the Subgraph-wise Sampling mechanism can also be treated as target-dependent,
    however, each set of target nodes for Subgraph-wise Sampling is determined by the sampled graph.

    Parameters
    ------------
    sampled_edges_for_layers:
        A sequence of tuple denoted as
        `( edge_index_for_sampled_graph, edge_id_in_integral_graph, (optional)edge_weight )`,
        where the `edge_index_for_sampled_graph` represents the sampled `edge_index` for sampled subgraph,
        the `edge_id_in_integral_graph` represents
        the corresponding positional indexes for the `edge_index` of integral graph,
        and the optional `edge_weight` for aggregation can also be provided.
    target_nodes_indexes:
        A tuple consists of (`torch.Tensor`, `torch.Tensor`),
        in which the first element represents the indexes of target nodes in sampled subgraph,
        and the second element represents the indexes of target nodes in the integral graph.
    all_sampled_nodes_indexes:
        Indexes of all sampled nodes for mini-batch.

    Attributes
    ------------
    target_nodes_indexes:
        A combined aggregation composed of
        `indexes_in_sampled_graph` and `indexes_in_integral_graph`
    all_sampled_nodes_indexes:
        Indexes of all sampled nodes for mini-batch.
    sampled_edges_for_layers:
        The stored sequence of tuple
        `( edge_index_for_sampled_graph, edge_id_in_integral_graph, (optional)edge_weight )`.
    """

    class _LayerSampledEdgeData:
        def __init__(
            self,
            edge_index_for_sampled_graph: torch.Tensor,
            edge_id_in_integral_graph: torch.Tensor,
            edge_weight: _typing.Optional[torch.Tensor],
        ):
            self.__edge_index_for_sampled_graph: torch.Tensor = (
                edge_index_for_sampled_graph
            )
            self.__edge_id_in_integral_graph: torch.Tensor = edge_id_in_integral_graph
            self.__edge_weight: _typing.Optional[torch.Tensor] = edge_weight

        @property
        def edge_index_for_sampled_graph(self) -> torch.LongTensor:
            edge_index_for_sampled_graph: _typing.Any = (
                self.__edge_index_for_sampled_graph
            )
            return edge_index_for_sampled_graph

        @property
        def edge_id_in_integral_graph(self) -> torch.LongTensor:
            edge_id_in_integral_graph: _typing.Any = self.__edge_id_in_integral_graph
            return edge_id_in_integral_graph

        @property
        def edge_weight(self) -> _typing.Optional[torch.Tensor]:
            return self.__edge_weight

    class _TargetNodes:
        @property
        def indexes_in_sampled_graph(self) -> torch.LongTensor:
            indexes_in_sampled_graph: _typing.Any = self.__indexes_in_sampled_graph
            return indexes_in_sampled_graph

        @property
        def indexes_in_integral_graph(self) -> torch.LongTensor:
            indexes_in_integral_graph: _typing.Any = self.__indexes_in_integral_graph
            return indexes_in_integral_graph

        def __init__(
            self,
            indexes_in_sampled_graph: torch.Tensor,
            indexes_in_integral_graph: torch.Tensor,
        ):
            self.__indexes_in_sampled_graph: torch.Tensor = indexes_in_sampled_graph
            self.__indexes_in_integral_graph: torch.Tensor = indexes_in_integral_graph

    @property
    def target_nodes_indexes(self) -> _TargetNodes:
        """ indexes of target nodes in the integral graph """
        return self.__target_nodes_indexes

    @property
    def all_sampled_nodes_indexes(self) -> torch.LongTensor:
        """ indexes of all sampled nodes in the integral graph """
        all_sampled_nodes_indexes: _typing.Any = self.__all_sampled_nodes_indexes
        return all_sampled_nodes_indexes

    @property
    def sampled_edges_for_layers(self) -> _typing.Sequence[_LayerSampledEdgeData]:
        return self.__sampled_edges_for_layers

    def __init__(
        self,
        sampled_edges_for_layers: _typing.Sequence[
            _typing.Tuple[torch.Tensor, torch.Tensor, _typing.Optional[torch.Tensor]]
        ],
        target_nodes_indexes: _typing.Tuple[torch.Tensor, torch.Tensor],
        all_sampled_nodes_indexes: torch.Tensor,
    ):
        self.__sampled_edges_for_layers: _typing.Sequence[
            TargetDependantSampledData._LayerSampledEdgeData
        ] = [
            self._LayerSampledEdgeData(item[0], item[1], item[2])
            for item in sampled_edges_for_layers
        ]
        self.__target_nodes_indexes: TargetDependantSampledData._TargetNodes = (
            self._TargetNodes(target_nodes_indexes[0], target_nodes_indexes[1])
        )
        self.__all_sampled_nodes_indexes: torch.Tensor = all_sampled_nodes_indexes


class TargetDependantSampler(torch.utils.data.DataLoader, _typing.Iterable):
    """
    An abstract base class for various target-dependent sampler
    """

    @classmethod
    def create_basic_sampler(
        cls,
        edge_index: torch.LongTensor,
        target_nodes_indexes: torch.LongTensor,
        layer_wise_arguments: _typing.Sequence,
        batch_size: int = 1,
        num_workers: int = 0,
        shuffle: bool = True,
        *args,
        **kwargs
    ) -> "TargetDependantSampler":
        """
        :param edge_index: edge index of integral graph
        :param target_nodes_indexes: indexes of target nodes in the integral graph
        :param layer_wise_arguments: layer-wise arguments for sampling
        :param batch_size: batch size for target nodes, default to 1
        :param num_workers: number of workers, default to 0
        :param shuffle: flag for shuffling, default to True
        :param args: remaining positional arguments
        :param kwargs: remaining keyword arguments
        :return: instance of TargetDependantSampler
        """
        raise NotImplementedError

    def __iter__(self):
        return super(TargetDependantSampler, self).__iter__()


class BasicLayerWiseTargetDependantSampler(TargetDependantSampler):
    """
    The base class for various Layer-wise Sampling techniques,
    providing basic functionality of composing sampled data for mini-batches.

    Parameters
    ------------
    edge_index:
        edge index of integral graph
    target_nodes_indexes:
        indexes of target nodes in the integral graph
    layer_wise_arguments:
        layer-wise arguments for sampling
    batch_size:
        batch size for target nodes
    num_workers:
        number of workers
    shuffle:
        flag for shuffling, default to True
    kwargs:
        remaining keyword arguments
    """

    def __init__(
        self,
        edge_index: torch.LongTensor,
        target_nodes_indexes: torch.LongTensor,
        layer_wise_arguments: _typing.Sequence,
        batch_size: _typing.Optional[int] = 1,
        num_workers: int = 0,
        shuffle: bool = True,
        **kwargs
    ):
        self._edge_index: torch.LongTensor = edge_index
        self.__layer_wise_arguments: _typing.Sequence = layer_wise_arguments
        if "collate_fn" in kwargs:
            del kwargs["collate_fn"]
        super(BasicLayerWiseTargetDependantSampler, self).__init__(
            target_nodes_indexes.unique().numpy(),
            batch_size,
            shuffle,
            num_workers=num_workers,
            collate_fn=self._collate_fn,
            **kwargs
        )

    @classmethod
    def create_basic_sampler(
        cls,
        edge_index: torch.LongTensor,
        target_nodes_indexes: torch.LongTensor,
        layer_wise_arguments: _typing.Sequence,
        batch_size: int = 1,
        num_workers: int = 0,
        shuffle: bool = True,
        *args,
        **kwargs
    ) -> TargetDependantSampler:
        """
        :param edge_index: edge index of integral graph
        :param target_nodes_indexes: indexes of target nodes in the integral graph
        :param layer_wise_arguments: layer-wise arguments for sampling
        :param batch_size: batch size for target nodes
        :param num_workers: number of workers
        :param shuffle: flag for shuffling, default to True
        :param args: remaining positional arguments
        :param kwargs: remaining keyword arguments
        :return: instance of TargetDependantSampler
        """
        return BasicLayerWiseTargetDependantSampler(
            edge_index,
            target_nodes_indexes,
            layer_wise_arguments,
            batch_size,
            num_workers,
            shuffle,
            **kwargs
        )

    def _sample_edges_for_layer(
        self,
        __current_layer_target_nodes_indexes: torch.LongTensor,
        __top_layer_target_nodes_indexes: torch.LongTensor,
        layer_argument: _typing.Any,
        *args,
        **kwargs
    ) -> _typing.Tuple[torch.LongTensor, _typing.Optional[torch.Tensor]]:
        """
        Sample edges for one specific layer, expected to be implemented in subclass.

        Parameters
        ------------
        __current_layer_target_nodes_indexes:
            target nodes for current layer
        __top_layer_target_nodes_indexes:
            target nodes for top layer
        layer_argument:
            argument for current layer
        args:
            remaining positional arguments
        kwargs:
            remaining keyword arguments

        Returns
        --------
        edge_id_in_integral_graph:
            the corresponding positional indexes for the `edge_index` of integral graph
        edge_weight:
            the optional `edge_weight` for aggregation
        """
        raise NotImplementedError

    def _collate_fn(
        self, top_layer_target_nodes_indexes_list: _typing.List[int]
    ) -> TargetDependantSampledData:
        return self.__sample_layers(
            torch.tensor(top_layer_target_nodes_indexes_list).unique()
        )

    def __sample_layers(
        self, __top_layer_target_nodes_indexes: torch.LongTensor
    ) -> TargetDependantSampledData:
        sampled_edges_for_layers: _typing.List[
            _typing.Tuple[torch.LongTensor, _typing.Optional[torch.Tensor]]
        ] = list()
        __current_layer_target_nodes_indexes: torch.LongTensor = (
            __top_layer_target_nodes_indexes
        )
        " Reverse self.__layer_wise_arguments from bottom-up to top-down "
        for layer_argument in self.__layer_wise_arguments[::-1]:
            current_layer_result: _typing.Tuple[
                torch.LongTensor, _typing.Optional[torch.Tensor]
            ] = self._sample_edges_for_layer(
                __current_layer_target_nodes_indexes,
                __top_layer_target_nodes_indexes,
                layer_argument,
            )
            __source_nodes_indexes_for_current_layer: torch.Tensor = self._edge_index[
                0, current_layer_result[0]
            ]
            __current_layer_target_nodes_indexes: torch.LongTensor = (
                __source_nodes_indexes_for_current_layer.unique()
            )
            sampled_edges_for_layers.append(current_layer_result)
        """ Reverse sampled_edges_for_layers from top-down to bottom-up """
        sampled_edges_for_layers: _typing.Sequence[
            _typing.Tuple[torch.LongTensor, _typing.Optional[torch.Tensor]]
        ] = sampled_edges_for_layers[::-1]

        sampled_nodes_in_sub_graph: torch.LongTensor = torch.cat(
            [
                self._edge_index[:, current_layer_result[0]].reshape([-1])
                for current_layer_result in sampled_edges_for_layers
            ]
        ).unique()
        __sampled_nodes_in_sub_graph_mapping: _typing.Dict[int, int] = dict(
            list(
                zip(
                    sampled_nodes_in_sub_graph.tolist(),
                    range(sampled_nodes_in_sub_graph.size(0)),
                )
            )
        )

        __sampled_edge_index_for_layers_in_sub_graph: _typing.Sequence[torch.Tensor] = [
            torch.stack(
                [
                    torch.tensor(
                        [
                            __sampled_nodes_in_sub_graph_mapping.get(node_index)
                            for node_index in self._edge_index[
                                0, current_layer_result[0]
                            ].tolist()
                        ]
                    ),
                    torch.tensor(
                        [
                            __sampled_nodes_in_sub_graph_mapping.get(node_index)
                            for node_index in self._edge_index[
                                1, current_layer_result[0]
                            ].tolist()
                        ]
                    ),
                ]
            )
            for current_layer_result in sampled_edges_for_layers
        ]

        return TargetDependantSampledData(
            [
                (temp_tuple[0], temp_tuple[1][0], temp_tuple[1][1])
                for temp_tuple in zip(
                    __sampled_edge_index_for_layers_in_sub_graph,
                    sampled_edges_for_layers,
                )
            ],
            (
                torch.tensor(
                    [
                        __sampled_nodes_in_sub_graph_mapping.get(
                            current_target_node_index_in_integral_data
                        )
                        for current_target_node_index_in_integral_data in __top_layer_target_nodes_indexes.tolist()
                        if current_target_node_index_in_integral_data
                        in __sampled_nodes_in_sub_graph_mapping
                    ]
                ).long(),  # Remap
                torch.tensor(
                    [
                        current_target_node_index_in_integral_data
                        for current_target_node_index_in_integral_data in __top_layer_target_nodes_indexes.tolist()
                        if current_target_node_index_in_integral_data
                        in __sampled_nodes_in_sub_graph_mapping
                    ]
                ).long(),
            ),
            sampled_nodes_in_sub_graph,
        )



try:
    import dgl
except ModuleNotFoundError:
    dgl = None
else:
    from ._to_dgl_dataset import to_dgl_dataset
try:
    import torch_geometric
except ModuleNotFoundError:
    torch_geometric = None
else:
    from ._to_pyg_dataset import to_pyg_dataset



import torch_geometric


class GraphSAINTSamplerFactory:
    """
    A simple factory class for creating varieties of
    :class:`torch_geometric.data.GraphSAINTSampler`.
    There exists potential sampling performance issues for
    the implementation of :class:`torch_geometric.data.GraphSAINTEdgeSampler`
    provided by PyTorch Geometric. Considering that the ultimate performance of
    GraphSAINT Edge Sampler and GraphSAINT Random Walk Sampler are similar
    according to the original literature
    `"GraphSAINT: Graph Sampling Based Inductive Learning Method"
    <https://arxiv.org/abs/1907.04931>`_ which introduces the GraphSAINT approach,
    nevertheless, when the walk length for GraphSAINT Random Walk Sampler is specified as `2`,
    the Random walk operation is actually selecting edges.
    Therefore an effective implementation for GraphSAINT Edge Sampler is not very urgently needed.
    Meanwhile, the varieties of Subgraph-wise sampling is scheduled to be redesigned and refactored.
    With the aim of abstracting a unified sampling module for representative mainstream varieties of
    Node-wise Sampling, Layer-wise Sampling, and Subgraph-wise Sampling.
    """

    @classmethod
    def create_node_sampler(
        cls,
        data,
        num_graphs_per_epoch: int,
        node_budget: int,
        sample_coverage_factor: int = 50,
        **kwargs
    ) -> torch_geometric.data.GraphSAINTNodeSampler:
        """
        A simple static method for instantiating :class:`torch_geometric.data.GraphSAINTNodeSampler`
        with more explicit arguments.

        Arguments
        ------------
        data:
            The conventional data of integral graph for sampling.
        num_graphs_per_epoch:
            number of subgraphs to sampler per epoch.
        node_budget:
            budget of nodes to sample for one sampled subgraph.
        sample_coverage_factor:
            The average number of samples per node should be used to
            compute normalization statistics.
        **kwargs:
            Additional optional arguments of :class:`torch.utils.data.DataLoader`,
            including :obj:`batch_size` or :obj:`num_workers`.

        Returns
        --------
        Instance of :class:`torch_geometric.data.GraphSAINTNodeSampler`.
        """
        return torch_geometric.data.GraphSAINTNodeSampler(
            data,
            node_budget,
            num_graphs_per_epoch,
            sample_coverage_factor,
            log=False,
            **kwargs
        )

    @classmethod
    def create_edge_sampler(
        cls,
        data,
        num_graphs_per_epoch: int,
        edge_budget: int,
        sample_coverage_factor: int = 50,
        **kwargs
    ) -> torch_geometric.data.GraphSAINTEdgeSampler:
        """
        A simple static method for instantiating :class:`torch_geometric.data.GraphSAINTEdgeSampler`
        with more explicit arguments.

        Arguments
        ------------
        data:
            The conventional data of integral graph for sampling.
        num_graphs_per_epoch:
            number of subgraphs to sampler per epoch.
        edge_budget:
            budget of edges to sample for one sampled subgraph.
        sample_coverage_factor:
            The average number of samples per node should be used to
            compute normalization statistics.
        **kwargs:
            Additional optional arguments of :class:`torch.utils.data.DataLoader`,
            including :obj:`batch_size` or :obj:`num_workers`.

        Returns
        --------
        Instance of :class:`torch_geometric.data.GraphSAINTEdgeSampler`.
        """
        return torch_geometric.data.GraphSAINTEdgeSampler(
            data,
            edge_budget,
            num_graphs_per_epoch,
            sample_coverage_factor,
            log=False,
            **kwargs
        )

    @classmethod
    def create_random_walk_sampler(
        cls,
        data,
        num_graphs_per_epoch: int,
        num_walks: int,
        walk_length: int,
        sample_coverage_factor: int = 50,
        **kwargs
    ) -> torch_geometric.data.GraphSAINTRandomWalkSampler:
        """
        A simple static method for instantiating :class:`torch_geometric.data.GraphSAINTEdgeSampler`
        with more explicit arguments.

        Arguments
        ------------
        data:
            The conventional data of integral graph for sampling.
        num_graphs_per_epoch:
            number of subgraphs to sampler per epoch.
        num_walks:
            The number of random walks for sampling.
        walk_length:
            The length of each random walk.
        sample_coverage_factor:
            The average number of samples per node should be used to
            compute normalization statistics.
        **kwargs:
            Additional optional arguments of :class:`torch.utils.data.DataLoader`,
            including :obj:`batch_size` or :obj:`num_workers`.

        Returns
        --------
        Instance of :class:`torch_geometric.data.GraphSAINTEdgeSampler`.
        """
        return torch_geometric.data.GraphSAINTRandomWalkSampler(
            data,
            num_walks,
            walk_length,
            num_graphs_per_epoch,
            sample_coverage_factor,
            log=False,
            **kwargs
        )



from ._split_edges import split_edges
from ._general import (
    index_to_mask,
    random_splits_mask,
    random_splits_mask_class,
    graph_cross_validation,
    graph_random_splits,
    graph_get_split,
    set_fold,
)



from ..base import BaseFeature
import numpy as np
import torch
from .. import register_feature


@register_feature("graph")
class BaseGraph(BaseFeature):
    def __init__(self, data_t="np", multigraph=True, **kwargs):
        super(BaseGraph, self).__init__(
            data_t=data_t, multigraph=multigraph, subgraph=True, **kwargs
        )

    def _preprocess(self, data):
        if not hasattr(data, "gf") or data["gf"] is None:
            data.gf = torch.FloatTensor([[]])

    def _postprocess(self, data):
        pass



import typing as _typing

_D = _typing.TypeVar('_D')


class _Schema(_typing.MutableMapping[str, _typing.Any]):
    def __setitem__(self, key: str, value: _typing.Any) -> None:
        self.__data[key] = value

    def __delitem__(self, key: str) -> None:
        del self.__data[key]

    def __getitem__(self, key: str) -> _typing.Any:
        return self.__data[key]

    def __len__(self) -> int:
        return len(self.__data)

    def __iter__(self) -> _typing.Iterator[str]:
        return iter(self.__data)

    def __init__(self):
        self.__data: _typing.MutableMapping[str, _typing.Any] = {}
        self.__meta_paths: _typing.Optional[
            _typing.Iterable[_typing.Iterable[str]]
        ] = None

    @property
    def meta_paths(self) -> _typing.Optional[
        _typing.Iterable[_typing.Iterable[str]]
    ]:
        return self.__meta_paths

    @meta_paths.setter
    def meta_paths(
            self, meta_paths: _typing.Optional[
                _typing.Iterable[_typing.Iterable[str]]
            ]
    ):
        self.__meta_paths = meta_paths


class Dataset(_typing.Iterable[_D], _typing.Sized):
    def __len__(self) -> int:
        raise NotImplementedError

    def __iter__(self) -> _typing.Iterator[_D]:
        raise NotImplementedError

    def __getitem__(self, index: int) -> _D:
        raise NotImplementedError

    def __setitem__(self, index: int, data: _D):
        raise NotImplementedError

    @property
    def train_split(self) -> _typing.Optional[_typing.Iterable[_D]]:
        raise NotImplementedError

    @property
    def val_split(self) -> _typing.Optional[_typing.Iterable[_D]]:
        raise NotImplementedError

    @property
    def test_split(self) -> _typing.Optional[_typing.Iterable[_D]]:
        raise NotImplementedError

    @property
    def train_index(self) -> _typing.Optional[_typing.AbstractSet[int]]:
        raise NotImplementedError

    @property
    def val_index(self) -> _typing.Optional[_typing.AbstractSet[int]]:
        raise NotImplementedError

    @property
    def test_index(self) -> _typing.Optional[_typing.AbstractSet[int]]:
        raise NotImplementedError

    @train_index.setter
    def train_index(self, train_index: _typing.Optional[_typing.Iterable[int]]):
        raise NotImplementedError

    @val_index.setter
    def val_index(self, val_index: _typing.Optional[_typing.Iterable[int]]):
        raise NotImplementedError

    @test_index.setter
    def test_index(self, test_index: _typing.Optional[_typing.Iterable[int]]):
        raise NotImplementedError

    @property
    def schema(self) -> _Schema:
        raise NotImplementedError


class _FoldsContainer:
    def __init__(
            self,
            folds: _typing.Optional[_typing.Iterable[_typing.Tuple[_typing.Sequence[int], _typing.Sequence[int]]]] = ...
    ):
        self._folds: _typing.Optional[_typing.List[_typing.Tuple[_typing.Sequence[int], _typing.Sequence[int]]]] = (
            list(folds) if isinstance(folds, _typing.Iterable) else None
        )
        if self._folds is not None and len(self._folds) == 0:
            self._folds = None

    @property
    def folds(self) -> _typing.Optional[_typing.Sequence[_typing.Tuple[_typing.Sequence[int], _typing.Sequence[int]]]]:
        if self._folds is not None and len(self._folds) == 0:
            self._folds = None
        return self._folds

    @folds.setter
    def folds(self, folds: _typing.Optional[_typing.Iterable[_typing.Tuple[_typing.Sequence[int], _typing.Sequence[int]]]]):
        self._folds: _typing.Optional[_typing.List[_typing.Tuple[_typing.Sequence[int], _typing.Sequence[int]]]] = (
            list(folds) if isinstance(folds, _typing.Iterable) else None
        )
        if self._folds is not None and len(self._folds) == 0:
            self._folds = None


class _FoldView:
    def __init__(self, folds_container: _FoldsContainer, fold_index: int):
        self._folds_container: _FoldsContainer = folds_container
        self._fold_index: int = fold_index

    @property
    def train_index(self) -> _typing.Sequence[int]:
        return self._folds_container.folds[self._fold_index][0]

    @property
    def val_index(self) -> _typing.Sequence[int]:
        return self._folds_container.folds[self._fold_index][1]


class _FoldsView(_typing.Sequence[_FoldView]):
    def __init__(self, folds_container: _FoldsContainer):
        self._folds_container = folds_container

    def __len__(self) -> int:
        return (
            len(self._folds_container.folds)
            if self._folds_container.folds is not None
            else 0
        )

    def __getitem__(self, fold_index: int) -> _FoldView:
        return _FoldView(self._folds_container, fold_index)


class InMemoryDataset(Dataset[_D]):
    @property
    def schema(self) -> _Schema:
        return self.__schema

    def __init__(
            self, data: _typing.Iterable[_D],
            train_index: _typing.Optional[_typing.Iterable[int]] = ...,
            val_index: _typing.Optional[_typing.Iterable[int]] = ...,
            test_index: _typing.Optional[_typing.Iterable[int]] = ...,
            schema: _typing.Optional[_Schema] = ...
    ):
        self.__data: _typing.MutableSequence[_D] = list(data)
        self.__train_index: _typing.Optional[_typing.Iterable[int]] = (
            train_index if isinstance(train_index, _typing.Iterable) else None
        )
        self.__val_index: _typing.Optional[_typing.Iterable[int]] = (
            val_index if isinstance(val_index, _typing.Iterable) else None
        )
        self.__test_index: _typing.Optional[_typing.Iterable[int]] = (
            test_index if isinstance(test_index, _typing.Iterable) else None
        )
        self.__schema: _Schema = schema if isinstance(schema, _Schema) else _Schema()
        self.__folds_container: _FoldsContainer = _FoldsContainer()

    @property
    def folds(self) -> _typing.Optional[_FoldsView]:
        return (
            _FoldsView(self.__folds_container)
            if (
                    self.__folds_container.folds is not None and
                    len(self.__folds_container.folds) > 0
            )
            else None
        )

    @folds.setter
    def folds(
            self,
            folds: _typing.Optional[
                _typing.Iterable[
                    _typing.Tuple[_typing.Sequence[int], _typing.Sequence[int]]
                ]
            ] = ...
    ):
        self.__folds_container.folds = folds

    def __len__(self) -> int:
        return len(self.__data)

    def __iter__(self) -> _typing.Iterator[_D]:
        return iter(self.__data)

    def __getitem__(self, index: int) -> _D:
        return self.__data[index]

    def __setitem__(self, index: int, data: _D):
        self.__data[index] = data

    @property
    def train_split(self) -> _typing.Optional[_typing.Iterable[_D]]:
        return (
            [self.__data[i] for i in self.__train_index]
            if isinstance(self.__train_index, _typing.Iterable) else None
        )

    @property
    def val_split(self) -> _typing.Optional[_typing.Iterable[_D]]:
        return (
            [self.__data[i] for i in self.__val_index]
            if isinstance(self.__val_index, _typing.Iterable) else None
        )

    @property
    def test_split(self) -> _typing.Optional[_typing.Iterable[_D]]:
        return (
            [self.__data[i] for i in self.__test_index]
            if isinstance(self.__test_index, _typing.Iterable) else None
        )

    @property
    def train_index(self) -> _typing.Optional[_typing.AbstractSet[int]]:
        return self.__train_index

    @property
    def val_index(self) -> _typing.Optional[_typing.AbstractSet[int]]:
        return self.__val_index

    @property
    def test_index(self) -> _typing.Optional[_typing.AbstractSet[int]]:
        return self.__test_index

    @train_index.setter
    def train_index(self, train_index: _typing.Optional[_typing.Iterable[int]]):
        if not (train_index is None or isinstance(train_index, _typing.Iterable)):
            raise TypeError
        elif train_index is None:
            self.__train_index: _typing.Optional[_typing.Iterable[int]] = None
        elif isinstance(train_index, _typing.Iterable):
            if len(list(train_index)) == 0:
                self.__train_index: _typing.Optional[_typing.Iterable[int]] = None
                return
            if not all([isinstance(i, int) for i in train_index]):
                raise TypeError
            if not (0 <= min(train_index) <= max(train_index) < len(self)):
                raise ValueError
            self.__train_index: _typing.Optional[_typing.Iterable[int]] = train_index

    @val_index.setter
    def val_index(self, val_index: _typing.Optional[_typing.Iterable[int]]):
        if not (val_index is None or isinstance(val_index, _typing.Iterable)):
            raise TypeError
        elif val_index is None:
            self.__val_index: _typing.Optional[_typing.Iterable[int]] = None
        elif isinstance(val_index, _typing.Iterable):
            if len(list(val_index)) == 0:
                self.__val_index: _typing.Optional[_typing.Iterable[int]] = None
                return
            if not all([isinstance(i, int) for i in val_index]):
                raise TypeError
            if not (0 <= min(val_index) <= max(val_index) < len(self)):
                raise ValueError
            self.__val_index: _typing.Optional[_typing.Iterable[int]] = val_index

    @test_index.setter
    def test_index(self, test_index: _typing.Optional[_typing.Iterable[int]]):
        if not (test_index is None or isinstance(test_index, _typing.Iterable)):
            raise TypeError
        elif test_index is None:
            self.__test_index: _typing.Optional[_typing.Set[int]] = None
        elif isinstance(test_index, _typing.Iterable):
            if len(list(test_index)) == 0:
                self.__test_index: _typing.Optional[_typing.Iterable[int]] = None
                return
            if not all([isinstance(i, int) for i in test_index]):
                raise TypeError
            if not (0 <= min(test_index) <= max(test_index) < len(self)):
                raise ValueError
            self.__test_index: _typing.Optional[_typing.Iterable[int]] = test_index



import torch
import torch.nn.functional as F
from typing import Sequence, Optional, Union, Tuple
from numbers import Real

from dgl.nn.pytorch.conv import GraphConv
from dgl import remove_self_loop, add_self_loop
import autogl.data
from . import register_model
from .base import BaseAutoModel, activate_func, ClassificationSupportedSequentialModel
from ....utils import get_logger


LOGGER = get_logger("GCNModel")


class GCN(torch.nn.Module):
    def __init__(self, args):
        super(GCN, self).__init__()
        self.args = args
        self.num_layer = int(self.args["num_layers"])

        missing_keys = list(
            set(
                [
                    "features_num",
                    "num_class",
                    "num_layers",
                    "hidden",
                    "dropout",
                    "act",
                ]
            )
            - set(self.args.keys())
        )
        if len(missing_keys) > 0:
            raise Exception("Missing keys: %s." % ",".join(missing_keys))

        if not self.num_layer == len(self.args["hidden"]) + 1:
            LOGGER.warn("Warning: layer size does not match the length of hidden units")
        self.convs = torch.nn.ModuleList()

        
        self.convs.append(
            GraphConv(
                self.args["features_num"],
                self.args["hidden"][0]
            )
        )

        for i in range(self.num_layer - 2):
            self.convs.append(
                GraphConv(
                    self.args["hidden"][i],
                    self.args["hidden"][i + 1]
                )
            )
        self.convs.append(
            GraphConv(
                self.args["hidden"][-1],
                self.args["num_class"]
            )
        )

    def forward(self, data):
        x = data.ndata['feat']
        for i in range(len(self.convs)):
            if i!=0:
                x = F.dropout(x, p=self.args["dropout"], training=self.training)
            x = self.convs[i](data, x)

            if i != self.num_layer - 1:
                x = activate_func(x, self.args["act"])

        return F.log_softmax(x, dim=1)


    def cls_encode(self, data) -> torch.Tensor:
        return self(data)

    def cls_decode(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.log_softmax(x, dim=1)

    def lp_encode(self, data):
        # discard the last layer, only use the layer before

        x = data.ndata['feat']
        for i in range(len(self.convs) - 1):
            if i != 0:
                x = F.dropout(x, p=self.args["dropout"], training=self.training)
            x = self.convs[i](data, x)

            if i != len(self.convs) - 2:
                x = activate_func(x, self.args["act"])

        return x

    def lp_decode(self, z, pos_edge_index, neg_edge_index):
        edge_index = torch.cat([pos_edge_index, neg_edge_index], dim=-1)
        logits = (z[edge_index[0]] * z[edge_index[1]]).sum(dim=-1)
        return logits

    def lp_decode_all(self, z):
        prob_adj = z @ z.t()
        return (prob_adj > 0).nonzero(as_tuple=False).t()


@register_model("gcn-model")
class AutoGCN(BaseAutoModel):
    r"""
    AutoGCN.
    The model used in this automodel is GCN, i.e., the graph convolutional network from the
    `"Semi-supervised Classification with Graph Convolutional
    Networks" <https://arxiv.org/abs/1609.02907>`_ paper. The layer is

    .. math::

        \mathbf{X}^{\prime} = \mathbf{\hat{D}}^{-1/2} \mathbf{\hat{A}}
        \mathbf{\hat{D}}^{-1/2} \mathbf{X} \mathbf{\Theta},

    where :math:`\mathbf{\hat{A}} = \mathbf{A} + \mathbf{I}` denotes the
    adjacency matrix with inserted self-loops and
    :math:`\hat{D}_{ii} = \sum_{j=0} \hat{A}_{ij}` its diagonal degree matrix.

    Parameters
    ----------
    num_features: ``int``
        The dimension of features.

    num_classes: ``int``
        The number of classes.

    device: ``torch.device`` or ``str``
        The device where model will be running on.

    init: `bool`.
        If True(False), the model will (not) be initialized.
    """

    def __init__(
        self,
        input_dimension: Optional[int] = None,
        output_dimension: Optional[int] = None,
        device: Union[str, torch.device] = 'cpu',
        **kwargs
    ) -> None:
        super().__init__(input_dimension, output_dimension, device, **kwargs)
        
        self.hyper_parameter_space = [
            {
                "parameterName": "add_self_loops",
                "type": "CATEGORICAL",
                "feasiblePoints": [1],
            },
            {
                "parameterName": "normalize",
                "type": "CATEGORICAL",
                "feasiblePoints": [1],
            },
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [128, 128, 128],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
        ]

        # initial point of hp search
        self.hyper_parameters = {
            "num_layers": 3,
            "hidden": [128, 64],
            "dropout": 0.,
            "act": "relu",
        }

    def _initialize(self):
        self._model = GCN({
            "features_num": self.input_dimension,
            "num_class": self.output_dimension,
            **self.hyper_parameters
        }).to(self.device)



"""
HPO Module for tuning hyper parameters
"""

import time
import numpy as np
from tqdm import trange
from . import register_hpo
from .base import BaseHPOptimizer, TimeTooLimitedError
import random
from .autone_file import utils

from ..feature import NetLSD as SgNetLSD

from autogl.backend import DependentBackend

_isdgl=DependentBackend.is_dgl()
if _isdgl:
    import dgl
else:
    from torch_geometric.data import InMemoryDataset
    from torch_geometric.data import GraphSAINTRandomWalkSampler

    class _MyDataset(InMemoryDataset):
        def __init__(self, datalist) -> None:
            super().__init__()
            self.data, self.slices = self.collate(datalist)

@register_hpo("autone")
class AutoNE(BaseHPOptimizer):
    """
    AutoNE HPOptimizer
    The Implementation of "AutoNE: Hyperparameter Optimization for Massive Network Embedding"(KDD 2019).
    See https://github.com/tadpole/AutoNE for more information

    Attributes
    ----------
    max_evals : int
        The max rounds of evaluating HPs
    subgraphs : int
        The number of subgraphs
    sub_evals : int
        The number of evaluation times on each subgraph
    sample_batch_size, sample_walk_length : int
        Using for sampling subgraph, see torch_geometric.data.GraphSAINRandomWalkSampler
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_evals = kwargs.get("max_evals", 100)
        self.subgraphs = kwargs.get("subgraphs", 2)
        self.sub_evals = kwargs.get("sub_evals", 2)
        self.sample_batch_size = kwargs.get("sample_batch_size", 150)
        self.sample_walk_length = kwargs.get("sample_walk_length", 100)

    def optimize(self, trainer, dataset, time_limit=None, memory_limit=None):
        """
        Optimize the HP by the method within give model and HP space

        See .base.BaseHPOptimizer.optimize
        """
        self.feval_name = trainer.get_feval(return_major=True).get_eval_name()
        self.is_higher_better = trainer.get_feval(return_major=True).is_higher_better()
        space = trainer.combined_hyper_parameter_space()
        # space = (
        #     trainer.hyper_parameter_space + trainer.get_model().hyper_parameter_space
        # )
        current_space = self._encode_para(space)

        def sample_subgraph(whole_data):
            data = whole_data.data
            loader = GraphSAINTRandomWalkSampler(
                data,
                batch_size=self.sample_batch_size,
                walk_length=self.sample_walk_length,
                num_steps=self.subgraphs,
                save_dir=whole_data.processed_dir,
            )
            results = []
            for data in loader:
                in_dataset = _MyDataset([data])
                results.append(in_dataset)
            return results

        def sample_subgraph_dgl(whole_data):
            data = whole_data[0] # dgl data
            # find data with different labels
            # random walk
            start = [random.randint(0, data.num_nodes - 1) for i in range(self.subgraphs)]
            traces, _  = dgl.sampling.random_walk_with_restart(data, start, length=self.sample_batch_size, restart_prob= 1 / self.sample_walk_length)
            subgraphs = dgl.node_subgraph(data, [traces[i, :] for i in traces.size(0)])
            return subgraphs

        func = SgNetLSD()

        def get_wne(graph):
            graph = func.fit_transform(graph)
            # transform = nx.NxGraph.compose(map(lambda x: x(), nx.NX_EXTRACTORS))
            # print(type(graph))
            # gf = transform.fit_transform(graph).data.gf
            gf = graph.data.gf
            fin = list(gf[0]) + list(map(lambda x: float(x), gf[1:]))
            return fin

        start_time = time.time()

        def fn(dset, para):
            para['dropout'] = float(para['dropout'])
            current_trainer = trainer.duplicate_from_hyper_parameter(para)
            current_trainer.train(dset)
            loss, self.is_higher_better = current_trainer.get_valid_score(dset)
            if self.is_higher_better:
                loss = -loss
            return current_trainer, loss

        # code in AutoNE
        sampled_number = self.subgraphs
        k = self.sub_evals
        s = self.max_evals
        X = []
        y = []
        params = utils.Params()
        params.set_space(current_space)
        total_t = 0.0
        info = []
        K = utils.K(len(params.type_))
        gp = utils.GaussianProcessRegressor(K)
        if _isdgl:
            sample_graphs = sample_subgraph_dgl(dataset)
        else:
            sample_graphs = sample_subgraph(dataset)
        print("Sample Phase:\n")
        for t in trange(sampled_number):
            b_t = time.time()
            i = t
            subgraph = sample_graphs[t]
            wne = get_wne(subgraph)
            for v in range(k):
                kargs = params.random_x()
                para = params.x2dict(kargs)
                externel_para, trial_para = self._decode_para(para)
                _, res = fn(subgraph, externel_para)
                X_reg = params.dict2x(trial_para)
                X.append(np.hstack((X_reg, wne)))
                y.append(res)

        best_res = None
        best_trainer = None
        best_para = None
        wne = get_wne(dataset)
        print("HPO Search Phase:\n")
        for t in trange(s):
            if time.time() - start_time > time_limit:
                self.logger.info("Time out of limit, Epoch: {}".format(str(i)))
                break
            b_t = time.time()
            gp.fit(np.vstack(X), y)
            X_temp, _ = gp.predict(params.get_bound(), params.get_type(), wne)
            X_temp = X_temp[: len(params.type_)]
            para = params.x2dict(X_temp)
            externel_para, trial_para = self._decode_para(para)
            current_trainer, res_temp = fn(dataset, externel_para)
            self._print_info(externel_para, res_temp)
            X_reg = params.dict2x(trial_para)

            X.append(np.hstack((X_reg, wne)))
            y.append(res_temp)
            if not best_res or res_temp < best_res:
                best_res = res_temp
                best_trainer = current_trainer
                best_para = para
            else:
                del current_trainer
            e_t = time.time()
            total_t += e_t - b_t

        if not best_res:
            raise TimeTooLimitedError(
                "Given time is too limited to finish one round in HPO."
            )

        decoded_json, _ = self._decode_para(best_para)
        self.logger.info("Best Parameter:")
        self._print_info(decoded_json, best_res)

        return best_trainer, decoded_json

    @classmethod
    def build_hpo_from_args(cls, args):
        """Build a new hpo instance."""
        raise NotImplementedError(
            "HP Optimizer must implement the build_hpo_from_args method"
        )



import typing as _typing
import torch.nn.functional
import torch_geometric
from torch_geometric.nn.conv import GINConv
from .. import base_encoder, encoder_registry
from ... import _utils


class _GIN(torch.nn.Module):
    def __init__(
            self, input_dimension: int,
            dimensions: _typing.Sequence[int],
            _act: str, _dropout: float,
            mlp_layers: int, _eps: str
    ):
        super(_GIN, self).__init__()

        self._act: str = _act

        def _get_act() -> torch.nn.Module:
            if _act == 'leaky_relu':
                return torch.nn.LeakyReLU()
            elif _act == 'relu':
                return torch.nn.ReLU()
            elif _act == 'elu':
                return torch.nn.ELU()
            elif _act == 'tanh':
                return torch.nn.Tanh()
            elif _act == 'PReLU'.lower():
                return torch.nn.PReLU()
            else:
                return torch.nn.ReLU()

        convolution_layers: torch.nn.ModuleList = torch.nn.ModuleList()
        batch_normalizations: torch.nn.ModuleList = torch.nn.ModuleList()

        __mlp_layers = [torch.nn.Linear(input_dimension, dimensions[0])]
        for _ in range(mlp_layers - 1):
            __mlp_layers.append(_get_act())
            __mlp_layers.append(torch.nn.Linear(dimensions[0], dimensions[0]))
        convolution_layers.append(
            GINConv(torch.nn.Sequential(*__mlp_layers), train_eps=_eps == "True")
        )
        batch_normalizations.append(torch.nn.BatchNorm1d(dimensions[0]))

        num_layers: int = len(dimensions)
        for layer in range(num_layers - 1):
            __mlp_layers = [torch.nn.Linear(dimensions[layer], dimensions[layer + 1])]
            for _ in range(mlp_layers - 1):
                __mlp_layers.append(_get_act())
                __mlp_layers.append(
                    torch.nn.Linear(dimensions[layer + 1], dimensions[layer + 1])
                )
            convolution_layers.append(
                GINConv(torch.nn.Sequential(*__mlp_layers), train_eps=_eps == "True")
            )
            batch_normalizations.append(
                torch.nn.BatchNorm1d(dimensions[layer + 1])
            )

        self.__convolution_layers: torch.nn.ModuleList = convolution_layers
        self.__batch_normalizations: torch.nn.ModuleList = batch_normalizations

    def forward(
            self, data: torch_geometric.data.Data, *__args, **__kwargs
    ) -> _typing.Sequence[torch.Tensor]:
        x: torch.Tensor = data.x
        edge_index: torch.Tensor = data.edge_index

        results: _typing.MutableSequence[torch.Tensor] = [x]
        num_layers = len(self.__convolution_layers)
        for layer in range(num_layers):
            x: torch.Tensor = self.__convolution_layers[layer](x, edge_index)
            x: torch.Tensor = _utils.activation.activation_func(x, self._act)
            x: torch.Tensor = self.__batch_normalizations[layer](x)
            results.append(x)
        return results


@encoder_registry.EncoderUniversalRegistry.register_encoder('gin')
@encoder_registry.EncoderUniversalRegistry.register_encoder('gin-encoder')
class GINEncoderMaintainer(base_encoder.AutoHomogeneousEncoderMaintainer):
    def _initialize(self) -> _typing.Optional[bool]:
        dimensions = list(self.hyper_parameters['hidden'])
        if (
                self.final_dimension not in (Ellipsis, None)
                and isinstance(self.final_dimension, int)
                and self.final_dimension > 0
        ):
            dimensions.append(self.final_dimension)
        self._encoder = _GIN(
            self.input_dimension, dimensions,
            self.hyper_parameters['act'],
            self.hyper_parameters['dropout'],
            self.hyper_parameters['mlp_layers'],
            self.hyper_parameters['eps']
        ).to(self.device)
        return True

    def __init__(
            self,
            input_dimension: _typing.Optional[int] = ...,
            final_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(GINEncoderMaintainer, self).__init__(
            input_dimension, final_dimension, device, *args, **kwargs
        )
        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "4,5,6",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 5,
                "minValue": [8, 8, 8, 8, 8],
                "maxValue": [64, 64, 64, 64, 64],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.9,
                "minValue": 0.1,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
            {
                "parameterName": "eps",
                "type": "CATEGORICAL",
                "feasiblePoints": ["True", "False"],
            },
            {
                "parameterName": "mlp_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
        ]
        self.hyper_parameters = {
            "num_layers": 3,
            "hidden": [64, 32],
            "dropout": 0.5,
            "act": "relu",
            "eps": "True",
            "mlp_layers": 2,
        }



import numpy as np
import torch
import typing
import autogl.data.graph
from .._feature_engineer import FeatureEngineer
from ..._data_preprocessor_registry import DataPreprocessorUniversalRegistry


class BaseFeatureSelector(FeatureEngineer):
    def __init__(self):
        super(BaseFeatureSelector, self).__init__()
        self._selection: typing.Optional[torch.Tensor] = None

    def __transform_homogeneous_static_graph(
            self, static_graph: autogl.data.graph.GeneralStaticGraph
    ) -> autogl.data.graph.GeneralStaticGraph:
        if (
                'x' in static_graph.nodes.data and
                isinstance(self._selection, (torch.Tensor, np.ndarray))
        ):
            static_graph.nodes.data['x'] = static_graph.nodes.data['x'][:, self._selection]
        if (
                'feat' in static_graph.nodes.data and
                isinstance(self._selection, (torch.Tensor, np.ndarray))
        ):
            static_graph.nodes.data['feat'] = static_graph.nodes.data['feat'][:, self._selection]
        return static_graph

    def _transform(
            self, data: typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]:
        if isinstance(data, autogl.data.graph.GeneralStaticGraph):
            return self.__transform_homogeneous_static_graph(data)
        elif (
                hasattr(data, 'x') and isinstance(data.x, torch.Tensor) and
                torch.is_tensor(data.x) and data.x.dim() == 2
        ):
            data.x = data.x[:, self._selection]
            return data
        else:
            return data


@DataPreprocessorUniversalRegistry.register_data_preprocessor("FilterConstant")
class FilterConstant(BaseFeatureSelector):
    r"""drop constant features"""

    def _fit(
            self, data: typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]
    ) -> typing.Union[autogl.data.graph.GeneralStaticGraph, typing.Any]:
        if isinstance(data, autogl.data.graph.GeneralStaticGraph):
            if 'x' in data.nodes.data:
                feature: typing.Optional[np.ndarray] = data.nodes.data['x'].numpy()
            elif 'feat' in data.nodes.data:
                feature: typing.Optional[np.ndarray] = data.nodes.data['feat'].numpy()
            else:
                feature: typing.Optional[np.ndarray] = None
        elif (
                hasattr(data, 'x') and isinstance(data.x, torch.Tensor) and
                torch.is_tensor(data.x) and data.x.dim() == 2
        ):
            feature: typing.Optional[np.ndarray] = data.x.numpy()
        else:
            feature: typing.Optional[np.ndarray] = None

        if feature is not None and isinstance(feature, np.ndarray) and feature.ndim == 2:
            self._selection: typing.Optional[torch.Tensor] = torch.from_numpy(
                np.where(np.all(feature == feature[0, :], axis=0) == np.array(False))[0]
            )
        else:
            self._selection: typing.Optional[torch.Tensor] = None
        return data



import importlib
import sys
from ...backend import DependentBackend
from . import _utils

from .decoders import (
    BaseDecoderMaintainer,
    DecoderUniversalRegistry,
    LogSoftmaxDecoderMaintainer,
    DotProductLinkPredictionDecoderMaintainer
)

from .encoders import (
    BaseEncoderMaintainer,
    AutoHomogeneousEncoderMaintainer,
    EncoderUniversalRegistry,
    GCNEncoderMaintainer,
    GATEncoderMaintainer,
    GINEncoderMaintainer,
    SAGEEncoderMaintainer
)

if DependentBackend.is_dgl():
    from .decoders import (
        TopKDecoderMaintainer,
        JKSumPoolDecoderMaintainer
    )
else:
    from .decoders import (
        DiffPoolDecoderMaintainer,
        SumPoolMLPDecoderMaintainer
    )

# load corresponding backend model of subclass
def _load_subclass_backend(backend):
    sub_module = importlib.import_module(f'.{backend.get_backend_name()}', __name__)
    this = sys.modules[__name__]
    for api, obj in sub_module.__dict__.items():
        setattr(this, api, obj)

_load_subclass_backend(DependentBackend)

__all__.extend([
    "BaseDecoderMaintainer",
    "DecoderUniversalRegistry",
    "LogSoftmaxDecoderMaintainer",
    "DotProductLinkPredictionDecoderMaintainer",
    "BaseEncoderMaintainer",
    "AutoHomogeneousEncoderMaintainer",
    "EncoderUniversalRegistry",
    "GCNEncoderMaintainer",
    "GATEncoderMaintainer",
    "GINEncoderMaintainer",
    "SAGEEncoderMaintainer"
])

if DependentBackend.is_dgl():
    __all__.extend([
        "TopKDecoderMaintainer",
        "JKSumPoolDecoderMaintainer",

    ])
else:
    __all__.extend([
        "DiffPoolDecoderMaintainer",
        "SumPoolMLPDecoderMaintainer"
    ])



import time
import numpy as np
import torch
import typing as _typing
from sklearn import preprocessing
from sklearn.metrics.pairwise import cosine_similarity
import tqdm
import tabulate
import autogl.data.graph
from ._feature_engineer import FeatureEngineer
from .._data_preprocessor_registry import DataPreprocessorUniversalRegistry
from ._selectors import GBDTFeatureSelector
from ....utils import get_logger

LOGGER = get_logger("Feature")


@DataPreprocessorUniversalRegistry.register_data_preprocessor("identity")
class IdentityFeature(FeatureEngineer):
    ...


@DataPreprocessorUniversalRegistry.register_data_preprocessor("OnlyConst".lower())
class OnlyConstFeature(FeatureEngineer):
    def _transform(
            self, data: _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[autogl.data.graph.GeneralStaticGraph, _typing.Any]:
        if isinstance(data, autogl.data.graph.GeneralStaticGraph):
            for node_t in data.nodes:
                for candidate_feature_key in ('feat', 'x'):
                    if candidate_feature_key in data.nodes[node_t].data:
                        data.nodes[node_t].data[candidate_feature_key] = torch.ones(
                            (data.nodes[node_t].data[candidate_feature_key].size(0), 1)
                        ).to(data.nodes[node_t].data[candidate_feature_key])
                    elif len(data.nodes[node_t].data) > 0:
                        _ref = data.nodes[node_t].data[list(data.nodes[node_t].data)[0]]
                        data.nodes[node_t].data[candidate_feature_key] = (
                            torch.ones((_ref.size(0), 1)).to(_ref)
                        )
                    else:
                        data.nodes[node_t].data[candidate_feature_key] = torch.ones(
                            (torch.unique(data.edges.connections).size(0), 1)
                        )
        elif hasattr(data, 'x') and isinstance(data.x, torch.Tensor):
            data.x = torch.ones((data.x.shape[0], 1)).to(data.x)
        elif hasattr(data, 'edge_index') and isinstance(data.edge_index, torch.Tensor):
            data.x = torch.ones((torch.unique(data.edge_index).size(0), 1)).to(data.edge_index)
        else:
            raise ValueError("Unsupported provided data")
        return data


def op_sum(x, nbs):
    res = np.zeros_like(x)
    for u in range(len(nbs)):
        nb = nbs[u]
        if len(nb != 0):
            res[u] = np.sum(x[nb], axis=0)
    return res


def op_mean(x, nbs):
    res = np.zeros_like(x)
    for u in range(len(nbs)):
        nb = nbs[u]
        if len(nb != 0):
            res[u] = np.mean(x[nb], axis=0)
    return res


def op_max(x, nbs):
    res = np.zeros_like(x)
    for u in range(len(nbs)):
        nb = nbs[u]
        if len(nb != 0):
            res[u] = np.max(x[nb], axis=0)
    return res


def op_min(x, nbs):
    res = np.zeros_like(x)
    for u in range(len(nbs)):
        nb = nbs[u]
        if len(nb != 0):
            res[u] = np.min(x[nb], axis=0)
    return res


def op_prod(x, nbs):
    res = np.zeros_like(x)
    for u in range(len(nbs)):
        nb = nbs[u]
        if len(nb != 0):
            res[u] = np.prod(x[nb], axis=0)
    return res


mms = preprocessing.MinMaxScaler()
ss = preprocessing.StandardScaler()


def scale(x):
    return ss.fit_transform(x)


class Timer:
    def __init__(self, timebudget=None):
        self._timebudget = timebudget
        self._esti_time = 0
        self._g_start = time.time()

    def start(self):
        self._start = time.time()

    def end(self):
        time_use = time.time() - self._start
        self._esti_time = (self._esti_time + time_use) / 2

    def is_timeout(self):
        timebudget = self._timebudget
        if timebudget:
            timebudget = self._timebudget - (time.time() - self._g_start)
            if timebudget < self._esti_time:
                return True
        return False


@DataPreprocessorUniversalRegistry.register_data_preprocessor('DeepGL'.lower())
class AutoFeatureEngineer(FeatureEngineer):
    r"""
    Notes
    -----
    An implementation of auto feature engineering method Deepgl [#]_ ,which iteratively generates features by aggregating neighbour features
    and select a fixed number of  features to automatically add important graph-aware features.
    References
    ----------
    .. [#] Rossi, R. A., Zhou, R., & Ahmed, N. K. (2020).
        Deep Inductive Graph Representation Learning.
        IEEE Transactions on Knowledge and Data Engineering, 32(3), 438–452.
        https://doi.org/10.1109/TKDE.2018.2878247
    Parameters
    ----------
    fix_length : int
        fixed number of features for every epoch. The final number of features added will be
        ``fixlen`` \times ``max_epoch``, 200 \times 5 in default.
    max_epoch : int
        number of epochs in total process.
    time_budget : int
        timebudget(seconds) for the feature engineering process, None for no time budget . Note that
        this time budget is a soft budget ,which is obtained by rough time estimation through previous iterations and
        may finally exceed the actual timebudget
    y_sel_func : Callable
        feature selector function object for selection at each iteration ,lightgbm in default. Note that in original paper,
        connected components of feature graph is used , and you may implement it by yourself if you want.
    verbosity : int
        hide any infomation except error and fatal if ``verbosity`` < 1
    """

    def __init__(
            self,
            fix_length: int = 200,
            max_epoch: int = 5,
            time_budget: _typing.Optional[int] = None,
            feature_selector=GBDTFeatureSelector,
            verbosity: int = 0,
            *args, **kwargs
    ):
        super(AutoFeatureEngineer, self).__init__()
        self._ops = [op_sum, op_mean, op_max, op_min]
        self._sim = cosine_similarity
        self._fixlen = fix_length
        self._max_epoch = max_epoch
        self._timebudget = time_budget
        self._feature_selector = feature_selector(
            fix_length, verbose_eval=verbosity >= 1, *args, **kwargs
        )
        self._verbosity = verbosity

    def _gen(self, x) -> np.ndarray:
        res = []
        for i, op in enumerate(self._ops):
            res.append(op(x, self.__neighbours))
        res = np.concatenate(res, axis=1)
        return res

    def _fit(self, homogeneous_static_graph: autogl.data.graph.GeneralStaticGraph):
        if not (
                homogeneous_static_graph.nodes.is_homogeneous and
                homogeneous_static_graph.edges.is_homogeneous
        ):
            raise ValueError
        if 'x' in homogeneous_static_graph.nodes.data:
            _feature_key = 'x'
            _original_features: torch.Tensor = (
                homogeneous_static_graph.nodes.data['x']
            )
        elif 'feat' in homogeneous_static_graph.nodes.data:
            _feature_key = 'feat'
            _original_features: torch.Tensor = (
                homogeneous_static_graph.nodes.data['feat']
            )
        else:
            raise ValueError

        num_nodes = _original_features.size(0)
        neighbours = [[] for _ in range(num_nodes)]
        for u, v in homogeneous_static_graph.edges.connections.t().numpy():
            neighbours[u].append(v)
        self.__neighbours: _typing.Sequence[np.ndarray] = tuple(
            [np.array(v) for v in neighbours]
        )

        x: np.ndarray = _original_features.numpy()
        gx: np.ndarray = x.copy()
        verbs = []
        soft_timer = Timer(self._timebudget)
        self._selection = []
        for epoch in tqdm.tqdm(range(self._max_epoch), disable=self._verbosity <= 0):
            soft_timer.start()
            verb = [epoch, gx.shape[1]]
            gx = self._gen(gx)
            gx = scale(gx)
            verb.append(gx.shape[1])

            homogeneous_static_graph.nodes.data[_feature_key] = torch.from_numpy(gx)
            self._feature_selector._fit(homogeneous_static_graph)
            self._selection.append(self._feature_selector._selection)
            homogeneous_static_graph = self._feature_selector._transform(
                homogeneous_static_graph
            )

            gx: np.ndarray = homogeneous_static_graph.nodes.data[_feature_key].numpy()
            verb.append(gx.shape[1])
            x = np.concatenate([x, gx], axis=1)
            verbs.append(verb)
            soft_timer.end()
            if soft_timer.is_timeout():
                break
        if self._verbosity >= 1:
            LOGGER.info(
                tabulate.tabulate(verbs, headers="epoch origin after-gen after-sel".split())
            )
        homogeneous_static_graph.nodes.data[_feature_key] = torch.from_numpy(x)
        return homogeneous_static_graph

    def _transform(self, homogeneous_static_graph: autogl.data.graph.GeneralStaticGraph):
        if not (
                homogeneous_static_graph.nodes.is_homogeneous and
                homogeneous_static_graph.edges.is_homogeneous
        ):
            raise ValueError
        if 'x' in homogeneous_static_graph.nodes.data:
            _feature_key = 'x'
            _original_features: torch.Tensor = (
                homogeneous_static_graph.nodes.data['x']
            )
        elif 'feat' in homogeneous_static_graph.nodes.data:
            _feature_key = 'feat'
            _original_features: torch.Tensor = (
                homogeneous_static_graph.nodes.data['feat']
            )
        else:
            raise ValueError

        x: np.ndarray = _original_features.numpy()
        gx: np.ndarray = x.copy()
        for selection in self._selection:
            gx = scale(self._gen(gx))[:, selection]
            x = np.concatenate([x, gx], axis=1)
        homogeneous_static_graph.nodes.data[_feature_key] = torch.from_numpy(x)
        return homogeneous_static_graph



import typing as _typing
from autogl.data import InMemoryDataset, Dataset
import autogl

if autogl.backend.DependentBackend.is_dgl():
    from ._dgl_compatible import split_edges_for_data
elif autogl.backend.DependentBackend.is_pyg():
    from ._pyg_compatible import split_edges_for_data
else:
    raise NotImplementedError


def split_edges(
        dataset: _typing.Iterable, train_ratio: float, val_ratio: _typing.Optional[float] = ...
) -> Dataset:
    if isinstance(val_ratio, float) and not 0 < train_ratio + val_ratio < 1:
        raise ValueError
    elif not 0 < train_ratio < 1:
        raise ValueError
    if (
            autogl.backend.DependentBackend.is_pyg() and
            not (isinstance(val_ratio, float) and 0 < val_ratio < 1)
    ):
        raise ValueError(
            "For PyG as backend, val_ratio MUST be specific float between 0 and 1, "
            "i.e. 0 < val_ratio < 1"
        )
    return (
        InMemoryDataset(
            [split_edges_for_data(item, train_ratio, val_ratio) for item in dataset],
            dataset.train_index, dataset.val_index, dataset.test_index, dataset.schema
        )
        if isinstance(dataset, Dataset)
        else
        InMemoryDataset(
            [split_edges_for_data(item, train_ratio, val_ratio) for item in dataset]
        )
    )



from ._base_feature_engineer import (
    BaseFeatureEngineer, BaseFeature
)
from ._feature_engineer_registry import (
    FeatureEngineerUniversalRegistry, FEATURE_DICT
)
from ._generators import (
    OneHotFeatureGenerator,
    EigenFeatureGenerator,
    GraphletGenerator,
    PageRankFeatureGenerator,
    LocalDegreeProfileGenerator,
    NormalizeFeatures,
    OneHotDegreeGenerator
)
from ._graph import (
    NetLSD,
    NXLargeCliqueSize,
    NXDegreeAssortativityCoefficient,
    NXDegreePearsonCorrelationCoefficient,
    NXHasBridges,
    NXGraphCliqueNumber,
    NXGraphNumberOfCliques,
    NXTransitivity,
    NXAverageClustering,
    NXIsConnected,
    NXNumberConnectedComponents,
    NXIsDistanceRegular,
    NXLocalEfficiency,
    NXGlobalEfficiency,
    NXIsEulerian,
)
from ._selectors import (
    FilterConstant, GBDTFeatureSelector
)
from ._auto_feature import (
    IdentityFeature, OnlyConstFeature, AutoFeatureEngineer
)

__all__ = [
    "BaseFeatureEngineer",
    "BaseFeature",
    "FeatureEngineerUniversalRegistry",
    "OneHotFeatureGenerator",
    "EigenFeatureGenerator",
    "GraphletGenerator",
    "PageRankFeatureGenerator",
    "LocalDegreeProfileGenerator",
    "NormalizeFeatures",
    "OneHotDegreeGenerator",
    "NetLSD",
    "NXLargeCliqueSize",
    "NXDegreeAssortativityCoefficient",
    "NXDegreePearsonCorrelationCoefficient",
    "NXHasBridges",
    "NXGraphCliqueNumber",
    "NXGraphNumberOfCliques",
    "NXTransitivity",
    "NXAverageClustering",
    "NXIsConnected",
    "NXNumberConnectedComponents",
    "NXIsDistanceRegular",
    "NXLocalEfficiency",
    "NXGlobalEfficiency",
    "NXIsEulerian",
    "FilterConstant",
    "GBDTFeatureSelector",
    "IdentityFeature",
    "OnlyConstFeature",
    "AutoFeatureEngineer"
]


"""
Base solver for classification problems
"""

from typing import Any
from ..base import BaseSolver
from ...module.ensemble import ENSEMBLE_DICT
from ...module import BaseEnsembler


class BaseClassifier(BaseSolver):
    """
    Base solver for classification problems
    """

    def predict_proba(self, *args, **kwargs) -> Any:
        """
        Predict the node probability.

        Returns
        -------
        result: Any
            The predicted probability
        """
        raise NotImplementedError()

    def set_ensemble_module(self, ensemble_module, *args, **kwargs) -> "BaseClassifier":
        """
        Set the ensemble module used in current solver.

        Parameters
        ----------
        ensemble_module: autogl.module.ensemble.BaseEnsembler or str or None
            The (name of) ensemble module used to ensemble the multi-models found.
            Disable ensemble by setting it to ``None``.

        Returns
        -------
        self: autogl.solver.BaseSolver
            A reference of current solver.
        """
        # load ensemble module
        if ensemble_module is None:
            self.ensemble_module = None
        elif isinstance(ensemble_module, BaseEnsembler):
            self.ensemble_module = ensemble_module
        elif isinstance(ensemble_module, str):
            if ensemble_module in ENSEMBLE_DICT:
                self.ensemble_module = ENSEMBLE_DICT[ensemble_module](*args, **kwargs)
            else:
                raise KeyError("cannot find ensemble module %s." % (ensemble_module))
        else:
            ValueError(
                "need ensemble module to be str or a BaseEnsembler instance, get",
                type(ensemble_module),
                "instead.",
            )



from .views_fn import (
    DropNode,
    PermuteEdge,
    MaskNode,
    SubGraph,
    RandomView
)

def get_view_by_name(view, aug_ratio):
    if view is None:
        return lambda x: x
    elif view == "dropN":
        return DropNode(aug_ratio=aug_ratio)
    elif view == "permE":
        return PermuteEdge(aug_ratio=aug_ratio)
    elif view == "subgraph":
        return SubGraph(aug_ratio=aug_ratio)
    elif view == "maskN":
        return MaskNode(aug_ratio=aug_ratio)
    elif view == "random2":
        canditates = [DropNode(aug_ratio=aug_ratio),
                      SubGraph(aug_ratio=aug_ratio)]
        return RandomView(candidates=canditates)
    elif view == "random3":
        canditates = [DropNode(aug_ratio=aug_ratio),
                      SubGraph(aug_ratio=aug_ratio),
                      PermuteEdge(aug_ratio=aug_ratio)]
        return RandomView(candidates=canditates)
    elif view == "random4":
        canditates = [DropNode(aug_ratio=aug_ratio),
                      SubGraph(aug_ratio=aug_ratio),
                      PermuteEdge(aug_ratio=aug_ratio),
                      MaskNode(aug_ratio=aug_ratio)]
        return RandomView(candidates=canditates)
    else:
        raise NotImplementedError(f'{view} is not supported yet. Support: ["dropN", "permE", "subgraph", \
                                    "maskN", "random2", "random3", "random4", None]')



import torch.nn.functional
import typing as _typing
import torch_geometric
from torch_geometric.nn.conv import GCNConv
from .. import base_encoder, encoder_registry
from ... import _utils


class _GCN(torch.nn.Module):
    def __init__(
            self, input_dimension: int, dimensions: _typing.Sequence[int],
            _act: _typing.Optional[str], _dropout: _typing.Optional[float]
    ):
        super(_GCN, self).__init__()
        self._act: _typing.Optional[str] = _act
        self._dropout: _typing.Optional[float] = _dropout
        self.__convolution_layers: torch.nn.ModuleList = torch.nn.ModuleList()
        for layer, output_dimension in enumerate(dimensions):
            self.__convolution_layers.append(
                GCNConv(input_dimension if layer == 0 else dimensions[layer - 1], output_dimension)
            )

    def forward(
            self, data: torch_geometric.data.Data, *__args, **__kwargs
    ) -> _typing.Sequence[torch.Tensor]:
        x: torch.Tensor = data.x
        edge_index: torch.LongTensor = data.edge_index
        if (
                hasattr(data, "edge_weight") and
                isinstance(getattr(data, "edge_weight"), torch.Tensor)
                and torch.is_tensor(data.edge_weight)
        ):
            edge_weight: _typing.Optional[torch.Tensor] = data.edge_weight
        else:
            edge_weight: _typing.Optional[torch.Tensor] = None
        results: _typing.MutableSequence[torch.Tensor] = [x]
        for layer, convolution_layer in enumerate(self.__convolution_layers):
            x = convolution_layer(x, edge_index, edge_weight)
            if layer < len(self.__convolution_layers) - 1:
                x: torch.Tensor = _utils.activation.activation_func(x, self._act)
                if isinstance(self._dropout, float) and 0 <= self._dropout <= 1:
                    x = torch.nn.functional.dropout(x, self._dropout, self.training)
            results.append(x)
        return results


@encoder_registry.EncoderUniversalRegistry.register_encoder('gcn')
@encoder_registry.EncoderUniversalRegistry.register_encoder('gcn_encoder')
class GCNEncoderMaintainer(base_encoder.AutoHomogeneousEncoderMaintainer):
    def _initialize(self) -> _typing.Optional[bool]:
        dimensions = list(self.hyper_parameters['hidden'])
        if (
                self.final_dimension not in (Ellipsis, None)
                and isinstance(self.final_dimension, int)
                and self.final_dimension > 0
        ):
            dimensions.append(self.final_dimension)
        self._encoder = _GCN(
            self.input_dimension, dimensions,
            self.hyper_parameters['act'], self.hyper_parameters['dropout']
        )
        return True

    def __init__(
            self,
            input_dimension: _typing.Optional[int] = ...,
            final_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(GCNEncoderMaintainer, self).__init__(
            input_dimension, final_dimension, device, *args, **kwargs
        )
        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [128, 128, 128],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
        ]
        self.hyper_parameters = {
            "num_layers": 2,
            "hidden": [16],
            "dropout": 0.2,
            "act": "leaky_relu",
        }



import typing
from autogl.utils import universal_registry
from . import _data_preprocessor


class DataPreprocessorUniversalRegistry(universal_registry.UniversalRegistryBase):
    @classmethod
    def register_data_preprocessor(cls, name: str) -> typing.Callable[
        [typing.Type[_data_preprocessor.DataPreprocessor]],
        typing.Type[_data_preprocessor.DataPreprocessor]
    ]:
        def register_data_preprocessor(
                data_preprocessor: typing.Type[_data_preprocessor.DataPreprocessor]
        ) -> typing.Type[_data_preprocessor.DataPreprocessor]:
            if not issubclass(data_preprocessor, _data_preprocessor.DataPreprocessor):
                raise TypeError
            else:
                cls[name] = data_preprocessor
                return data_preprocessor

        return register_data_preprocessor

    @classmethod
    def get_data_preprocessor(cls, name: str) -> typing.Type[_data_preprocessor.DataPreprocessor]:
        if name not in cls:
            raise ValueError(f"Data Preprocessor with name \"{name}\" not exist")
        else:
            return cls[name]



import torch
from deeprobust.graph.global_attack import Random,DICE
import numpy as np
import scipy.sparse as sp
from tqdm import tqdm
import time
import torch.nn.functional as F

from . import register_nas_estimator
from ..space import BaseSpace
from .base import BaseEstimator
from ..backend import *
from ...train.evaluation import Acc
from ..utils import get_hardware_aware_metric

@register_nas_estimator("grna")
class GRNAEstimator(BaseEstimator):
    """
    Graph robust neural architecture estimator under adversarial attack.

    Use model directly to get estimations.

    Parameters
    ----------
    loss_f : str
        The name of loss funciton in PyTorch
    evaluation : list of Evaluation
        The evaluation metrics in module/train/evaluation
        GRNA_metric = acc_metric+ robustness_metric
    lambda: float
        The hyper-parameter to balance the accuracy metric and robustness metric to perform ultimate evaluation
    perturb_type: str
        Perturbation methods to simulate the adversarial attack process
    adv_sample_num: int
        Adversarial sample number used in measure architecture robustness.
    """

    def __init__(self, 
                 loss_f="nll_loss", 
                 evaluation=[Acc()], 
                 lambda_=0.05, 
                 perturb_type='random',
                 adv_sample_num=10,  
                 dis_type='ce',
                 ptbr=0.05):
        super().__init__(loss_f, evaluation)
        self.evaluation = evaluation
        self.lambda_ = lambda_
        self.perturb_type = perturb_type
        self.dis_type = dis_type
        self.adv_sample_num = adv_sample_num
        self.ptbr = ptbr
        print('initialize GRNA estimator')

    def infer(self, model: BaseSpace, dataset, mask="train"):
        
        device = next(model.parameters()).device
        dset = dataset[0].to(device)
        mask = bk_mask(dset, mask)

        pred = model(dset)[mask]
        label = bk_label(dset)
        y = label[mask]

        loss = getattr(F, self.loss_f)(pred, y)
        probs = F.softmax(pred, dim=1).detach().cpu().numpy()

        # robustness metric
        dist = 0
        for _ in range(self.adv_sample_num):
            modified_adj = self.gen_adversarial_samples(dset.edge_index, num_nodes=dset.num_nodes, perturb_prop=self.ptbr, attack_method=self.perturb_type)
            d_data = dset.clone()
            d_data = d_data.to(device)
            edge_index = torch.LongTensor(np.vstack((modified_adj.tocoo().row,modified_adj.tocoo().col)))
            d_data.edge_index = edge_index.to(device)
            perturb_pred = model(d_data)[mask]
            dist += distance(perturb_pred, pred, dis_type=self.dis_type)
        dist = dist/self.adv_sample_num
        
        y = y.cpu()
        metrics = [eva.evaluate(probs, y)+self.lambda_*dist for eva in self.evaluation]

        return metrics, loss


    def gen_adversarial_samples(self, edge_index, num_nodes, perturb_prop, attack_method='random'):
    
        if num_nodes is None:
            num_nodes = max(edge_index[0])+1
        
        edge_index = edge_index.detach().cpu().numpy()
        delta = int(edge_index.shape[1]//2 * perturb_prop)
        v = np.ones_like(edge_index[0])
        adj = sp.csr_matrix((v,(edge_index[0], edge_index[1])), shape=(num_nodes,num_nodes))
        if attack_method=='random':
            attacker = Random()
            attacker.attack(adj, n_perturbations=delta, type='flip')
        elif attack_method=='dice':
            labels = self.data.y.cpu().numpy()
            attacker = DICE()
            attacker.attack(adj, labels, delta)
        else:
            assert False, 'Wrong Type of attack method!'

        modified_adj = attacker.modified_adj  # scipy.sparse matrix
        
        return modified_adj.tocsr()




def distance(perturb, clean, dis_type='ce', data=None, p=2):
    """
    Distance between logits of perturbed and clean data.
    Parameters:
    ---------
    perturb: torch.Tensor [n, C]
    clean: torch.Tensor [n,C]
    type: loss type
    labels: ground truth labels, needed when type='cw'
    p: fro norm, needed when type='fro'

    Return
    ------
    Distance: torch.Tensor [n,]
    """
    # if type=='cos':
    #     return perturb*clean / torch.sqrt(torch.norm(perturb,p=2) * torch.norm(clean))
    if dis_type=='fro':
        return torch.norm(perturb - clean,p=p)

    elif dis_type=='ce':
        p_ = F.softmax(clean,-1)
        logq = F.log_softmax(perturb, -1)
        return -(p_*logq).mean()

    elif dis_type=='kl':
        logq = F.log_softmax(perturb,-1)
        p_ = F.softmax(clean, -1)
        logp = F.log_softmax(clean,-1)
        return (p_*(logp-logq)).mean()*100

    elif dis_type=='cw':
        perturb, clean, labels = perturb[data.train_mask],clean[data.train_mask], data.y[data.train_mask]
        eye = torch.eye(labels.max() + 1)
        onehot_mx = eye[labels]
        one_hot_labels =  onehot_mx.to(labels.device)
        # perturb
        ptb_best_second_class = (perturb - 1000*one_hot_labels).argmax(1)
        margin = perturb[np.arange(len(perturb)), labels]  - \
            perturb[np.arange(len(perturb)), ptb_best_second_class]
        ptb_loss = -torch.clamp(margin, max = 0, min = -0.1).mean()
        # clean
        clean_best_second_class = (perturb - 1000*one_hot_labels).argmax(1)
        margin = clean[np.arange(len(perturb)), labels]  - \
            clean[np.arange(len(perturb)), clean_best_second_class]
        clean_loss = -torch.clamp(margin, max = 0, min = -0.1).mean()

        return (ptb_loss-clean_loss)*100


import torch.nn.functional
import autogl
from ._basic import BaseFeatureGenerator
from ._pyg_impl import degree, scatter_min, scatter_max, scatter_mean, scatter_std
from ..._data_preprocessor_registry import DataPreprocessorUniversalRegistry


@DataPreprocessorUniversalRegistry.register_data_preprocessor("LocalDegreeProfile")
class LocalDegreeProfileGenerator(BaseFeatureGenerator):
    r"""Appends the Local Degree Profile (LDP) from the `"A Simple yet
    Effective Baseline for Non-attribute Graph Classification"
    <https://arxiv.org/abs/1811.03508>`_ paper

    .. math::
        \mathbf{x}_i = \mathbf{x}_i \, \Vert \, (\deg(i), \min(DN(i)),
        \max(DN(i)), \textrm{mean}(DN(i)), \textrm{std}(DN(i)))

    to the node features, where :math:`DN(i) = \{ \deg(j) \mid j \in
    \mathcal{N}(i) \}`.
    """

    def _extract_nodes_feature(self, data: autogl.data.Data) -> torch.Tensor:
        row, col = data.edge_index
        if data.x is not None and isinstance(data.x, torch.Tensor):
            N = data.x.size(0)
        else:
            N = (torch.max(data.edge_index).item() + 1)

        deg = degree(row, N, dtype=torch.float)
        deg_col = deg[col]

        min_deg, _ = scatter_min(deg_col, row, dim_size=N)
        min_deg[min_deg > 10000] = 0
        max_deg, _ = scatter_max(deg_col, row, dim_size=N)
        max_deg[max_deg < -10000] = 0
        mean_deg = scatter_mean(deg_col, row, dim_size=N)
        std_deg = scatter_std(deg_col, row, dim_size=N)

        x = torch.stack([deg, min_deg, max_deg, mean_deg, std_deg], dim=1)
        return x


@DataPreprocessorUniversalRegistry.register_data_preprocessor("NormalizeFeatures")
class NormalizeFeatures(BaseFeatureGenerator):
    def __init__(self):
        super(NormalizeFeatures, self).__init__(override_features=True)

    def _extract_nodes_feature(self, data: autogl.data.Data) -> torch.Tensor:
        if data.x is not None and isinstance(data.x, torch.Tensor):
            data.x.div_(data.x.sum(dim=-1, keepdim=True).clamp_(min=1.))
        return data.x


@DataPreprocessorUniversalRegistry.register_data_preprocessor("OneHotDegree")
class OneHotDegreeGenerator(BaseFeatureGenerator):
    r"""Adds the node degree as one hot encodings to the node features.

    Args:
        max_degree (int): Maximum degree.
        in_degree (bool, optional): If set to :obj:`True`, will compute the
            in-degree of nodes instead of the out-degree.
            (default: :obj:`False`)
        cat (bool, optional): Concat node degrees to node features instead
            of replacing them. (default: :obj:`True`)
    """

    def __init__(
            self, max_degree: int = 1000,
            in_degree: bool = False, cat: bool = True
    ):
        self.__max_degree: int = max_degree
        self.__in_degree: bool = in_degree
        self.__cat: bool = cat
        super(OneHotDegreeGenerator, self).__init__()

    def _extract_nodes_feature(self, data: autogl.data.Data) -> torch.Tensor:
        idx, x = data.edge_index[1 if self.__in_degree else 0], data.x
        deg = degree(idx, data.num_nodes, dtype=torch.long)
        deg = torch.nn.functional.one_hot(
            deg, num_classes=self.__max_degree + 1
        ).to(torch.float)
        return deg



from torch_geometric.transforms.one_hot_degree import OneHotDegree
from torch_geometric.utils import degree
from torch_geometric.transforms.normalize_features import NormalizeFeatures
from torch_geometric.transforms.local_degree_profile import LocalDegreeProfile
from .. import register_feature
from .base import BaseGenerator
import numpy as np
from .. import register_feature
import torch
from functools import wraps

PYG_GENERATORS = []


def register_pyg(cls):
    PYG_GENERATORS.append(cls)
    register_feature(cls.__name__)(cls)
    return cls


class PYGGenerator(BaseGenerator):
    def __init__(self, *args, **kwargs):
        super(PYGGenerator, self).__init__(data_t="tensor")
        self._args = args
        self._kwargs = kwargs

    def extract(self, data):
        return data.x

    def _transform(self, data):
        dsc = self.extract(data)
        # data.x = torch.cat([data.x, dsc], dim=1)
        data.x = dsc
        return data


def pygfunc(func):
    r"""A decorator for pyg transforms. You may want to use it to quickly wrap a feature transform function object.
    Examples
    --------
    @register_pyg
    @pygfunc(local_degree_profile)
    class PYGLocalDegreeProfile(local_degree_profile):pass
    """

    def decorator_func(cls):
        cls.extract = lambda s, data: (func(*s._args, **s._kwargs)(data)).x
        return cls

    return decorator_func


@register_pyg
@pygfunc(LocalDegreeProfile)
class PYGLocalDegreeProfile(PYGGenerator):
    pass


# def _preprocess(self,data):
#     num_nodes=data.num_nodes
#     self._num_nodes=num_nodes
#     data.num_nodes=np.sum(num_nodes)
# def _postprocess(self,data):
#     data.num_nodes=self._num_nodes


@register_pyg
@pygfunc(NormalizeFeatures)
class PYGNormalizeFeatures(PYGGenerator):
    pass


@register_pyg
@pygfunc(OneHotDegree)
class PYGOneHotDegree(PYGGenerator):
    def __init__(self, max_degree=1000):
        super(PYGOneHotDegree, self).__init__(max_degree=max_degree)

    """
    def _transform(self, data):
        #idx, x = data.edge_index[0], data.x
        #deg = degree(idx, data.num_nodes, dtype=torch.long)
        #self._kwargs["max_degree"] = np.min(
        #    [self._kwargs["max_degree"], torch.max(deg).numpy()]
        #)
        dsc = self.extract(data)
        data.x = torch.cat([data.x, dsc], dim=1)
        return data
    """



"""
Utilities used by the solver

* LeaderBoard: The LeaderBoard that maintains the performance of models.
"""

import random
import typing as _typing
import torch
import torch.backends.cudnn
import numpy as np
import pandas as pd
from ..backend import DependentBackend
from ..data import Dataset
from ..data.graph import GeneralStaticGraph

from ..utils import get_logger
LOGGER = get_logger("LeaderBoard")

BACKEND = DependentBackend.get_backend_name()

if BACKEND == 'dgl':
    from autogl.datasets.utils.conversion import to_dgl_dataset as _convert_dataset
else:
    from autogl.datasets.utils.conversion import to_pyg_dataset as _convert_dataset

class LeaderBoard:
    """
    The leaderBoard that can be used to store / sort the model performance automatically.

    Parameters
    ----------
    fields: list of `str`
        A list of field name that shows the model performance. The first field is used as
        the major field for sorting the model performances.

    is_higher_better: `dict` of *field* -> `bool`
        A mapping of indicator that whether each field is higher better.
    """

    def __init__(self, fields, is_higher_better):
        assert isinstance(fields, list)
        self.keys = ["name"] + fields
        self.perform_dict = pd.DataFrame(columns=self.keys)
        self.is_higher_better = is_higher_better
        self.major_field = fields[0]

    def set_major_field(self, field) -> None:
        """
        Set the major field of current LeaderBoard.

        Parameters
        ----------
        field: `str`
            The major field, should be one of the fields when initialized.

        Returns
        -------
        None
        """
        if field in self.keys and not field == "name":
            self.major_field = field
        else:
            LOGGER.warning(
                f"Field [{field}] NOT found in the current LeaderBoard, will ignore."
            )

    def insert_model_performance(self, name, performance) -> None:
        """
        Add/Override a record of model performance. If name given is already in the leaderboard,
        will overrride the slot.

        Parameters
        ----------
        name: `str`
            The model name/identifier that identifies the model.

        performance: `dict`
            The performance dict. The key inside the dict should be the fields when initialized.
            The value of the dict should be the corresponding scores.

        Returns
        -------
        None
        """
        if name not in self.perform_dict["name"]:
            # we just add a new row
            performance["name"] = name
            new = pd.DataFrame(performance, index=[0])
            self.perform_dict = self.perform_dict.append(new, ignore_index=True)
        else:
            LOGGER.warning(
                "model already in the leaderboard, will override current result."
            )
            self.remove_model_performance(name)
            self.insert_model_performance(name, performance)

    def remove_model_performance(self, name) -> None:
        """
        Remove the record of given models.

        Parameters
        ----------
        name: `str`
            The model name/identifier that needed to be removed.

        Returns
        -------
        None
        """
        if name not in self.perform_dict["name"]:
            LOGGER.warning(
                "no model detected in current leaderboard, will ignore removing action."
            )
            return
        index = self.perform_dict["name"][self.perform_dict["name"] == name].index
        self.perform_dict.drop(self.perform_dict.index[index], inplace=True)
        return

    def get_best_model(self, index=0) -> str:
        """
        Get the best model according to the performance of the major field.

        Parameters
        ----------
        index: `int`
            The index of the model (from good to bad). Default `0`.

        Returns
        -------
        name: `str`
            The name/identifier of the required model.
        """
        sorted_df = self.perform_dict.sort_values(
            by=self.major_field, ascending=not self.is_higher_better[self.major_field]
        )
        name_list = sorted_df["name"].tolist()
        if "ensemble" in name_list:
            name_list.remove("ensemble")
        return name_list[index]

    def show(self, top_k=0) -> None:
        """
        Show current LeaderBoard (from best model to worst).

        Parameters
        ----------
        top_k: `int`
            Controls the number model shown.
            If less than or equal to `0`, will show all the models. Default to `0`.

        Returns
        -------
        None
        """
        top_k: int = top_k if top_k > 0 else len(self.perform_dict)

        """
        reindex self.__performance_data_frame
        to ensure the columns of name and representation are in left-side of the data frame
        """
        _columns = self.perform_dict.columns.tolist()
        maxcolwidths: _typing.List[_typing.Optional[int]] = []
        if "name" in _columns:
            _columns.remove("name")
            _columns.insert(0, "name")
            maxcolwidths.append(40)
        self.perform_dict = self.perform_dict[_columns]

        sorted_performance_df: pd.DataFrame = self.perform_dict.sort_values(
            self.major_field, ascending=not self.is_higher_better[self.major_field]
        )
        sorted_performance_df = sorted_performance_df.head(top_k)

        from tabulate import tabulate

        _columns = sorted_performance_df.columns.tolist()
        maxcolwidths.extend([None for _ in range(len(_columns) - len(maxcolwidths))])
        print(
            tabulate(
                list(zip(*[sorted_performance_df[column] for column in _columns])),
                headers=_columns,
                tablefmt="grid",
            )
        )

def get_graph_from_dataset(dataset, graph_id=0):
    if isinstance(dataset, Dataset):
        return dataset[graph_id]
    if BACKEND == 'pyg': return dataset[graph_id]
    if BACKEND == 'dgl':
        from dgl import DGLGraph
        data = dataset[graph_id]
        if isinstance(data, DGLGraph): return data
        return data[0]
    
def get_graph_node_number(graph):
    # FIXME: if the feature is None, this will throw an error
    if isinstance(graph, GeneralStaticGraph):
        if BACKEND == 'pyg':
            return graph.nodes.data['x'].size(0)
        return graph.nodes.data['feat'].size(0)
    if BACKEND == 'pyg':
        size = graph.x.shape[0]
    else:
        size = graph.num_nodes()
    return size

def get_graph_node_features(graph):
    if isinstance(graph, GeneralStaticGraph):
        if BACKEND == 'dgl' and 'feat' in graph.nodes.data:
            return graph.nodes.data['feat']
        if BACKEND == 'pyg' and 'x' in graph.nodes.data:
            return graph.nodes.data['x']
        return None
    if BACKEND == 'pyg' and hasattr(graph, 'x'):
        return graph.x
    elif BACKEND == 'dgl' and 'feat' in graph.ndata:
        return graph.ndata['feat']
    return None

def get_graph_masks(graph, mask='train'):
    if isinstance(graph, GeneralStaticGraph):
        if f'{mask}_mask' in graph.nodes.data:
            return graph.nodes.data[f'{mask}_mask']
        return None
    if BACKEND == 'pyg' and hasattr(graph, f'{mask}_mask'):
        return getattr(graph, f'{mask}_mask')
    if BACKEND == 'dgl' and f'{mask}_mask' in graph.ndata:
        return graph.ndata[f'{mask}_mask']
    return None

def get_graph_labels(graph):
    if isinstance(graph, GeneralStaticGraph):
        if 'label' in graph.nodes.data and BACKEND == 'dgl':
            return graph.nodes.data['label']
        if 'y' in graph.nodes.data and BACKEND == 'pyg':
            return graph.nodes.data['y']
        return None
    if BACKEND == 'pyg' and hasattr(graph, 'y'): return graph.y
    if BACKEND == 'dgl' and 'label' in graph.ndata: return graph.ndata['label']
    return None

def get_dataset_labels(dataset):
    if isinstance(dataset[0], GeneralStaticGraph):
        return torch.LongTensor([d.data['label' if BACKEND == 'dgl' else 'y'] for d in dataset])
    if BACKEND == 'pyg':
        return dataset.data.y
    else:
        return torch.LongTensor([d[1] for d in dataset])

def convert_dataset(dataset):
    # todo: replace the trick by re-implementing the convert_dataset in utils
    if hasattr(dataset[0], "edges"): return _convert_dataset(dataset)
    # if isinstance(dataset, Dataset): return _convert_dataset(dataset)
    return dataset

def set_seed(seed=None):
    """
    Set seed of whole process
    """
    if seed is None:
        seed = random.randint(0, 5000)

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

def get_graph_labels_hetero(graph, target_node_type):
    if isinstance(graph, GeneralStaticGraph):
        if 'label' in graph.nodes[target_node_type].data and BACKEND == 'dgl':
            return graph.nodes[target_node_type].data['label']
        return None
    if BACKEND == 'dgl' and 'label' in graph.ndata[target_node_type]: return graph.ndata[target_node_type]['label']



import itertools
import os

import scipy.io
import torch
import typing as _typing

from autogl.data import Data, download_url, InMemoryStaticGraphSet
from autogl.data.graph import GeneralStaticGraphGenerator
from ._dataset_registry import DatasetUniversalRegistry
from ._data_source import OnlineDataSource
from .. import backend as _backend


class _MATLABMatrix(OnlineDataSource):
    @property
    def _raw_filenames(self) -> _typing.Iterable[str]:
        splits = [self.__name]
        files = ["mat"]
        return [
            "{}.{}".format(s, f) for s, f
            in itertools.product(splits, files)
        ]

    @property
    def _processed_filenames(self) -> _typing.Iterable[str]:
        return ["data.pt"]

    def _fetch(self):
        for name in self._raw_filenames:
            download_url(self.__url + name, self._raw_directory)

    def _process(self):
        path = os.path.join(self._raw_directory, f"{self.__name}.mat")
        mat = scipy.io.loadmat(path)
        adj_matrix, group = mat["network"], mat["group"]

        y = torch.from_numpy(group.todense()).to(torch.float)

        row_ind, col_ind = adj_matrix.nonzero()
        edge_index = torch.stack([torch.tensor(row_ind), torch.tensor(col_ind)], dim=0)
        edge_attr = torch.tensor(adj_matrix[row_ind, col_ind])
        data = Data(edge_index=edge_index, edge_attr=edge_attr, x=None, y=y)
        torch.save(data, list(self._processed_file_paths)[0])

    def __len__(self) -> int:
        return 1

    def __getitem__(self, index: int):
        if index != 0:
            raise IndexError
        return self.__data

    def __init__(self, path: str, name: str, url: str):
        self.__name: str = name
        self.__url: str = url
        super(_MATLABMatrix, self).__init__(path)
        self.__data = torch.load(
            list(self._processed_file_paths)[0]
        )


@DatasetUniversalRegistry.register_dataset("BlogCatalog".lower())
class BlogCatalogDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        filename: str = "BlogCatalog".lower()
        url: str = "http://leitang.net/code/social-dimension/data/"
        data = _MATLABMatrix(path, filename, url)[0]
        if _backend.DependentBackend.is_dgl():
            super(BlogCatalogDataset, self).__init__(
                [
                    GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                        {'label': data.y}, data.edge_index,
                        {'edge_attr': data.edge_attr}
                    )
                ]
            )
        elif _backend.DependentBackend.is_pyg():
            super(BlogCatalogDataset, self).__init__(
                [
                    GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                        {'y': data.y}, data.edge_index,
                        {'edge_attr': data.edge_attr}
                    )
                ]
            )


@DatasetUniversalRegistry.register_dataset("WikiPEDIA".lower())
class WIKIPEDIADataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        filename: str = "POS"
        url = "http://snap.stanford.edu/node2vec/"
        data = _MATLABMatrix(path, filename, url)[0]
        if _backend.DependentBackend.is_dgl():
            super(WIKIPEDIADataset, self).__init__(
                [
                    GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                        {'label': data.y}, data.edge_index,
                        {'attr': data.edge_attr}
                    )
                ]
            )
        elif _backend.DependentBackend.is_pyg():
            super(WIKIPEDIADataset, self).__init__(
                [
                    GeneralStaticGraphGenerator.create_homogeneous_static_graph(
                        {'y': data.y}, data.edge_index,
                        {'attr': data.edge_attr}
                    )
                ]
            )



import copy
import torch
import typing as _typing
import dgl
from autogl.data.graph import GeneralStaticGraph
import autogl.data.graph.utils.conversion
from . import _base_feature_engineer


class BaseFeatureEngineer(
    _base_feature_engineer.BaseFeatureEngineer
):
    @classmethod
    def __preprocess(
            cls, data: _typing.Union[dgl.DGLGraph, GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[GeneralStaticGraph, _typing.Any]:
        if isinstance(data, dgl.DGLGraph):
            graph = autogl.data.graph.utils.conversion.dgl_graph_to_general_static_graph(data)
            setattr(graph, '_ORIGINAL_TYPE_BEFORE_FE', 'DGLGraph')
            return graph
        else:
            return data

    @classmethod
    def __postprocess(
            cls, data: _typing.Union[GeneralStaticGraph, _typing.Any]
    ) -> _typing.Union[dgl.DGLGraph, GeneralStaticGraph, _typing.Any]:
        if (
                isinstance(data, GeneralStaticGraph) and
                hasattr(data, '_ORIGINAL_TYPE_BEFORE_FE') and
                isinstance(getattr(data, '_ORIGINAL_TYPE_BEFORE_FE'), str) and
                getattr(data, '_ORIGINAL_TYPE_BEFORE_FE') == 'DGLGraph'
        ):
            return autogl.data.graph.utils.conversion.general_static_graph_to_dgl_graph(data)
        else:
            return data

    def fit(self, dataset):
        dataset = copy.deepcopy(dataset)
        with torch.no_grad():
            for i, data in enumerate(dataset):
                dataset[i] = self.__postprocess(
                    self._postprocess(self._fit(self._preprocess(self.__preprocess(data))))
                )
        return dataset

    def transform(self, dataset, inplace: bool = True):
        if not inplace:
            dataset = copy.deepcopy(dataset)
        with torch.no_grad():
            for i, data in enumerate(dataset):
                dataset[i] = self.__postprocess(
                    self._postprocess(self._transform(self._preprocess(self.__preprocess(data))))
                )
        return dataset



import torch
from deeprobust.graph.global_attack import Random,DICE
import numpy as np
import scipy.sparse as sp
from tqdm import tqdm
import time
import torch.nn.functional as F

from . import register_nas_estimator
from ..space import BaseSpace
from .base import BaseEstimator
from ..backend import *
from ...train.evaluation import Acc
from ..utils import get_hardware_aware_metric

@register_nas_estimator("oneshot")
class OneShotEstimator(BaseEstimator):
    """
    One shot estimator.

    Use model directly to get estimations.

    Parameters
    ----------
    loss_f : str
        The name of loss funciton in PyTorch
    evaluation : list of Evaluation
        The evaluation metrics in module/train/evaluation
    """

    def __init__(self, loss_f="nll_loss", evaluation=[Acc()]):
        super().__init__(loss_f, evaluation)
        self.evaluation = evaluation

    def infer(self, model: BaseSpace, dataset, mask="train"):
        device = next(model.parameters()).device
        dset = dataset[0].to(device)
        mask = bk_mask(dset, mask)

        pred = model(dset)[mask]
        label = bk_label(dset)
        y = label[mask]

        loss = getattr(F, self.loss_f)(pred, y)
        probs = F.softmax(pred, dim=1).detach().cpu().numpy()
        
        y = y.cpu()
        metrics = [eva.evaluate(probs, y) for eva in self.evaluation]
        return metrics, loss


@register_nas_estimator("oneshot_hardware")
class OneShotEstimator_HardwareAware(OneShotEstimator):
    """
    One shot hardware-aware estimator.

    Use model directly to get estimations with some hardware-aware metrics.

    Parameters
    ----------
    loss_f : str
        The name of loss funciton in PyTorch
    evaluation : list of Evaluation
        The evaluation metrics in module/train/evaluation
    hardware_evaluation : str or runable
        The hardware-aware metrics. Can be 'parameter' or 'latency'. Or you can define a special metric by a runable function
    hardware_metric_weight : float
        The weight of hardware-aware metric, which will be a bias added to metrics
    """

    def __init__(
        self,
        loss_f="nll_loss",
        evaluation=[Acc()],
        hardware_evaluation="parameter",
        hardware_metric_weight=0,
    ):
        super().__init__(loss_f, evaluation)
        self.hardware_evaluation = hardware_evaluation
        self.hardware_metric_weight = hardware_metric_weight

    def infer(self, model: BaseSpace, dataset, mask="train"):
        metrics, loss = super().infer(model, dataset, mask)
        if isinstance(self.hardware_evaluation, str):
            hardware_metric = get_hardware_aware_metric(model, self.hardware_evaluation)
        else:
            hardware_metric = self.hardware_evaluation(model)
        metrics = [x - hardware_metric * self.hardware_metric_weight for x in metrics]
        metrics.append(hardware_metric)
        return metrics, loss





from .data import Data
from ._dataset import Dataset, InMemoryDataset, InMemoryStaticGraphSet
from .download import download_url
from .extract import extract_tar, extract_zip, extract_bz2, extract_gz

__all__ = [
    "Data",
    "Dataset",
    "InMemoryDataset",
    "InMemoryStaticGraphSet",
    "download_url",
    "extract_tar",
    "extract_zip",
    "extract_bz2",
    "extract_gz",
]



import torch.nn.functional
import typing as _typing
import torch_geometric
from torch_geometric.nn.conv import GraphConv
from torch_geometric.nn.pool import TopKPooling
from torch_geometric.nn.glob import (
    global_add_pool, global_max_pool, global_mean_pool
)
from ...encoders import base_encoder
from .. import base_decoder, decoder_registry
from ... import _utils


class _LogSoftmaxDecoder(torch.nn.Module):
    def forward(self, features: _typing.Sequence[torch.Tensor], *__args, **__kwargs) -> torch.Tensor:
        return torch.nn.functional.log_softmax(features[-1], dim=1)


@decoder_registry.DecoderUniversalRegistry.register_decoder('log_softmax')
@decoder_registry.DecoderUniversalRegistry.register_decoder('log_softmax_decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('LogSoftmax'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('LogSoftmax_decoder'.lower())
class LogSoftmaxDecoderMaintainer(base_decoder.BaseDecoderMaintainer):
    def _initialize(self, *args, **kwargs) -> _typing.Optional[bool]:
        self._decoder = _LogSoftmaxDecoder().to(self.device)
        return True


class _SumPoolMLPDecoder(torch.nn.Module):
    def __init__(
            self, _final_dimension: int, hidden_dimension: int, output_dimension: int,
            _act: _typing.Optional[str], _dropout: _typing.Optional[float],
            num_graph_features: _typing.Optional[int]
    ):
        super(_SumPoolMLPDecoder, self).__init__()
        if (
                isinstance(num_graph_features, int)
                and num_graph_features > 0
        ):
            _final_dimension += num_graph_features
            self.__num_graph_features: _typing.Optional[int] = num_graph_features
        else:
            self.__num_graph_features: _typing.Optional[int] = None
        self._fc1: torch.nn.Linear = torch.nn.Linear(
            _final_dimension, hidden_dimension
        )
        self._fc2: torch.nn.Linear = torch.nn.Linear(
            hidden_dimension, output_dimension
        )
        self._act: _typing.Optional[str] = _act
        self._dropout: _typing.Optional[float] = _dropout

    def forward(
            self, features: _typing.Sequence[torch.Tensor],
            data: torch_geometric.data.Data, *__args, **__kwargs
    ):
        feature = features[-1]
        feature = global_add_pool(feature, data.batch)
        if (
                isinstance(self.__num_graph_features, int)
                and self.__num_graph_features > 0
        ):
            if (
                    hasattr(data, 'gf') and
                    isinstance(data.gf, torch.Tensor) and data.gf.dim() == 2 and
                    data.gf.size() == (feature.size(0), self.__num_graph_features)
            ):
                graph_features: torch.Tensor = data.gf
            else:
                raise ValueError(
                    f"The provided data is expected to contain property 'gf' "
                    f"with {self.__num_graph_features} dimensions as graph feature"
                )
            feature: torch.Tensor = torch.cat([feature, graph_features], dim=-1)
        feature: torch.Tensor = self._fc1(feature)
        feature: torch.Tensor = _utils.activation.activation_func(feature, self._act)
        if isinstance(self._dropout, float) and 0 <= self._dropout <= 1:
            feature: torch.Tensor = torch.nn.functional.dropout(
                feature, self._dropout, self.training
            )
        feature: torch.Tensor = self._fc2(feature)
        return torch.nn.functional.log_softmax(feature, dim=-1)


@decoder_registry.DecoderUniversalRegistry.register_decoder('SumPoolMLP'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('SumPoolMLPDecoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('SumPoolMLP_Decoder'.lower())
class SumPoolMLPDecoderMaintainer(base_decoder.BaseDecoderMaintainer):
    def _initialize(self, encoder: base_encoder.AutoHomogeneousEncoderMaintainer, *args, **kwargs) -> _typing.Optional[bool]:
        if (
                isinstance(getattr(self, "num_graph_features"), int) and
                getattr(self, "num_graph_features") > 0
        ):
            num_graph_features: _typing.Optional[int] = getattr(self, "num_graph_features")
        else:
            num_graph_features: _typing.Optional[int] = None
        self._decoder = _SumPoolMLPDecoder(
            tuple(encoder.get_output_dimensions())[-1],
            self.hyper_parameters['hidden'], self.output_dimension,
            self.hyper_parameters['act'], self.hyper_parameters['dropout'],
            num_graph_features
        ).to(self.device)
        return True

    def __init__(
            self, output_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(SumPoolMLPDecoderMaintainer, self).__init__(
            output_dimension, device, *args, **kwargs
        )
        self.num_graph_features = kwargs.get("num_graph_features", 0)
        self.hyper_parameter_space = [
            {
                "parameterName": "hidden",
                "type": "INTEGER",
                "maxValue": 64,
                "minValue": 8,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.9,
                "minValue": 0.1,
                "scalingType": "LINEAR",
            }
        ]
        self.hyper_parameters = {
            "hidden": 32,
            "act": "relu",
            "dropout": 0.5
        }


class _DiffPoolDecoder(torch.nn.Module):
    def __init__(
            self, input_dimension: int, output_dimension: int,
            _ratio: _typing.Union[float, int], _dropout: _typing.Optional[float],
            _act: _typing.Optional[str], num_graph_features: _typing.Optional[int]
    ):
        super(_DiffPoolDecoder, self).__init__()
        self.input_dimension = input_dimension
        self.output_dimension = output_dimension
        self.ratio: _typing.Union[float, int] = _ratio
        self._act: _typing.Optional[str] = _act
        self.dropout: _typing.Optional[float] = _dropout
        self.num_graph_features: _typing.Optional[int] = num_graph_features

        self.conv1 = GraphConv(self.input_dimension, 128)
        self.pool1 = TopKPooling(128, ratio=self.ratio)
        self.conv2 = GraphConv(128, 128)
        self.pool2 = TopKPooling(128, ratio=self.ratio)
        self.conv3 = GraphConv(128, 128)
        self.pool3 = TopKPooling(128, ratio=self.ratio)

        if (
                isinstance(self.num_graph_features, int)
                and self.num_graph_features > 0
        ):
            self.lin1 = torch.nn.Linear(256 + self.num_graph_features, 128)
        else:
            self.lin1 = torch.nn.Linear(256, 128)
        self.lin2 = torch.nn.Linear(128, 64)
        self.lin3 = torch.nn.Linear(64, self.output_dimension)

    def forward(
            self, features: _typing.Sequence[torch.Tensor],
            data: torch_geometric.data.Data, *__args, **__kwargs
    ):
        x: torch.Tensor = features[-1]
        edge_index: torch.LongTensor = data.edge_index
        batch = data.batch
        if (
                self.num_graph_features is not None and
                isinstance(self.num_graph_features, int) and
                self.num_graph_features > 0
        ):
            if not (
                    hasattr(data, 'gf') and
                    isinstance(data.gf, torch.Tensor) and
                    data.gf.size() == (x.size(0), self.num_graph_features)
            ):
                raise ValueError(
                    f"The provided data is expected to contain property 'gf' "
                    f"with {self.num_graph_features} dimensions as graph feature"
                )

        x = torch.nn.functional.relu(self.conv1(x, edge_index))
        x, edge_index, _, batch, _, _ = self.pool1(x, edge_index, None, batch)
        x1 = torch.cat([global_max_pool(x, batch), global_mean_pool(x, batch)], dim=1)

        x = torch.nn.functional.relu(self.conv2(x, edge_index))
        x, edge_index, _, batch, _, _ = self.pool2(x, edge_index, None, batch)
        x2 = torch.cat([global_max_pool(x, batch), global_mean_pool(x, batch)], dim=1)

        x = torch.nn.functional.relu(self.conv3(x, edge_index))
        x, edge_index, _, batch, _, _ = self.pool3(x, edge_index, None, batch)
        x3 = torch.cat([global_max_pool(x, batch), global_mean_pool(x, batch)], dim=1)

        x = x1 + x2 + x3
        if (
                isinstance(self.num_graph_features, int)
                and self.num_graph_features > 0
        ):
            x = torch.cat([x, data.gf], dim=-1)
        x = self.lin1(x)
        x = _utils.activation.activation_func(x, self._act)
        x = torch.nn.functional.dropout(x, p=self.dropout, training=self.training)
        x = self.lin2(x)
        x = _utils.activation.activation_func(x, self._act)
        x = torch.nn.functional.log_softmax(self.lin3(x), dim=-1)
        return x


@decoder_registry.DecoderUniversalRegistry.register_decoder('DiffPool'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('DiffPoolDecoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('DiffPool_decoder'.lower())
class DiffPoolDecoderMaintainer(base_decoder.BaseDecoderMaintainer):
    def _initialize(
            self, encoder: base_encoder.AutoHomogeneousEncoderMaintainer, *args, **kwargs
    ) -> _typing.Optional[bool]:
        if (
                isinstance(getattr(self, "num_graph_features"), int) and
                getattr(self, "num_graph_features") > 0
        ):
            num_graph_features: _typing.Optional[int] = getattr(self, "num_graph_features")
        else:
            num_graph_features: _typing.Optional[int] = None
        self._decoder = _DiffPoolDecoder(
            list(encoder.get_output_dimensions())[-1],
            self.output_dimension,
            self.hyper_parameters['ratio'], self.hyper_parameters['dropout'],
            self.hyper_parameters['act'], num_graph_features
        ).to(self.device)
        return True

    def __init__(
            self, output_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(DiffPoolDecoderMaintainer, self).__init__(
            output_dimension, device, *args, **kwargs
        )
        self.num_graph_features = kwargs.get("num_graph_features", 0)
        self.hyper_parameter_space = [
            {
                "parameterName": "ratio",
                "type": "DOUBLE",
                "maxValue": 0.9,
                "minValue": 0.1,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.9,
                "minValue": 0.1,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
        ]
        self.hyper_parameters = {
            "ratio": 0.8,
            "dropout": 0.5,
            "act": "relu"
        }

class _DotProductLinkPredictonDecoder(torch.nn.Module):
    def forward(self,
        features: _typing.Sequence[torch.Tensor],
        graph: torch_geometric.data.Data,
        pos_edge: torch.Tensor,
        neg_edge: torch.Tensor,
        **__kwargs
    ):
        z = features[-1]
        edge_index = torch.cat([pos_edge, neg_edge], dim=-1)
        logits = (z[edge_index[0]] * z[edge_index[1]]).sum(dim=-1)
        return logits

@decoder_registry.DecoderUniversalRegistry.register_decoder('lpdecoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('dotproduct'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('lp-decoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('dot-product'.lower())
class DotProductLinkPredictionDecoderMaintainer(base_decoder.BaseDecoderMaintainer):
    def _initialize(self, *args, **kwargs):
        self._decoder = _DotProductLinkPredictonDecoder()



import torch
import typing as _typing
from ...hpo import AutoModule


class BaseEncoderMaintainer(AutoModule):
    def __init__(
            self,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(BaseEncoderMaintainer, self).__init__(
            device, *args, **kwargs
        )
        self._encoder: _typing.Optional[torch.nn.Module] = None

    @property
    def encoder(self) -> _typing.Optional[torch.nn.Module]:
        return self._encoder

    def to_device(self, device: _typing.Union[torch.device, str, int, None]):
        self.device = device
        if (
                self._encoder not in (Ellipsis, None) and
                isinstance(self._encoder, torch.nn.Module)
        ):
            self._encoder.to(self.device)

    def from_hyper_parameter(
            self, hyper_parameter: _typing.Mapping[str, _typing.Any], **kwargs
    ):
        raise NotImplementedError

    def _initialize(self) -> _typing.Optional[bool]:
        raise NotImplementedError


class AutoHomogeneousEncoderMaintainer(BaseEncoderMaintainer):
    def _initialize(self) -> _typing.Optional[bool]:
        raise NotImplementedError

    def __init__(
            self,
            input_dimension: _typing.Optional[int] = ...,
            final_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        self._input_dimension: _typing.Optional[int] = input_dimension
        self._final_dimension: _typing.Optional[int] = final_dimension
        super(AutoHomogeneousEncoderMaintainer, self).__init__(
            device, *args, **kwargs
        )
        self.__args: _typing.Tuple[_typing.Any, ...] = args
        self.__kwargs: _typing.Mapping[str, _typing.Any] = kwargs

    @property
    def input_dimension(self) -> _typing.Optional[int]:
        return self._input_dimension
    
    @input_dimension.setter
    def input_dimension(self, input_dimension):
        self._input_dimension = input_dimension

    @property
    def final_dimension(self):
        return self._final_dimension
    
    @final_dimension.setter
    def final_dimension(self, final_dimension):
        # TODO: may mutate search space according to the final dimension
        self._final_dimension = final_dimension

    def from_hyper_parameter(
            self, hyper_parameter: _typing.Mapping[str, _typing.Any], **kwargs
    ):
        new_kwargs = dict(self.__kwargs)
        new_kwargs.update(kwargs)
        duplicate: AutoHomogeneousEncoderMaintainer = self.__class__(
            self.input_dimension, self.final_dimension, self.device,
            **new_kwargs
        )
        hp = dict(self.hyper_parameters)
        hp.update(hyper_parameter)
        duplicate.hyper_parameters = hp
        duplicate.initialize()
        return duplicate

    def get_output_dimensions(self) -> _typing.Iterable[int]:
        """"""
        ''' Note that this is a default implicit assumption '''
        _output_dimensions = [self._input_dimension]
        _output_dimensions.extend(self.hyper_parameters["hidden"])
        if (
                isinstance(self.final_dimension, int) and
                self.final_dimension > 0
        ):
            _output_dimensions.append(self.final_dimension)
        return _output_dimensions



"""
Auto solver for various graph tasks
"""

from autogl.backend import DependentBackend
from .classifier import AutoGraphClassifier, AutoNodeClassifier, AutoLinkPredictor

if DependentBackend.is_dgl():
    from .classifier import AutoHeteroNodeClassifier

from .utils import LeaderBoard

__all__ = [
    "AutoNodeClassifier",
    "AutoGraphClassifier",
    "AutoLinkPredictor",
    "LeaderBoard",
]

if DependentBackend.is_dgl():
    __all__.append("AutoHeteroNodeClassifier")


import typing as _typing


class CanonicalEdgeType(_typing.Sequence[str]):
    def __init__(self, source_node_type: str, relation_type: str, target_node_type: str):
        if not isinstance(source_node_type, str):
            raise TypeError
        elif ' ' in source_node_type:
            raise ValueError
        if not isinstance(relation_type, str):
            raise TypeError
        elif ' ' in relation_type:
            raise ValueError
        if not isinstance(target_node_type, str):
            raise TypeError
        elif ' ' in target_node_type:
            raise ValueError
        self.__source_node_type: str = source_node_type
        self.__relation_type: str = relation_type
        self.__destination_node_type: str = target_node_type

    @property
    def source_node_type(self) -> str:
        return self.__source_node_type

    @property
    def relation_type(self) -> str:
        return self.__relation_type

    @property
    def target_node_type(self) -> str:
        return self.__destination_node_type

    def __eq__(self, other):
        if not (isinstance(other, CanonicalEdgeType) or isinstance(other, _typing.Sequence)):
            return False
        elif isinstance(other, _typing.Sequence):
            if not (len(other) == 3 and all([(isinstance(t, str) and ' ' not in t) for t in other])):
                raise TypeError
            return (
                    other[0] == self.source_node_type and
                    other[1] == self.relation_type and
                    other[2] == self.target_node_type
            )
        elif isinstance(other, CanonicalEdgeType):
            return (
                    other.source_node_type == self.source_node_type and
                    other.relation_type == self.relation_type and
                    other.target_node_type == self.target_node_type
            )

    def __getitem__(self, index: int):
        return (self.source_node_type, self.relation_type, self.target_node_type)[index]

    def __len__(self) -> int:
        return 3
